import requests
import time 
from random import randint
from threading import Thread
from bs4 import BeautifulSoup as bs
from dataclasses import dataclass
import logging
import csv
import sys
import baangt.base.GlobalConstants as GC
from dataclasses_json import dataclass_json


logger = logging.getLogger("pyC")


@dataclass_json
@dataclass
class proxy_data:
    ip: str = ""
    port: str = ""
    username: str = ""
    password: str = ""
    type: str = "http"
    called: int = 0
    failed: int = 0

    def __post_init__(self):
        if not isinstance(self.called, int):
            self.called = int(self.failed)
        if not isinstance(self.failed, int):
            self.failed = int(self.failed)

    def Called(self):
        self.called+=1

    def Failed(self):
        self.failed+=1

    def Reset(self):
        self.called = 0
        self.failed = 0


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class ProxyRotate(metaclass=Singleton):
    """
    In ProxyRotate class we combine all functionalities related to proxy server handling.

    If the class is called with "reReadProxies"=True, we'll scrape and test new proxy servers
    If the class is called without reReadProxies we'll read existing proxies from the list.

    There are 3 states, that a proxy can have:
    * New (we just gathered this proxy from a page. We don't know whether it works or not.
    * Failed (we tried this proxy before. It didn't work)
    * Working (Fail-Count < GlobalConstants.Proxy_failcount)

    External interface are methods:
      random_proxy to retrieve a proxy
      remove_proxy to report an error
      recheckProxies to scrape for new proxies

    """

    def __init__(self, reReadProxies=True):
        self.reReadProxies = reReadProxies
        self.proxy_gather_link = "https://www.sslproxies.org/"
        self.proxies = {}                     # Currently usable proxies
        self.all_proxies = {}                 # List of all Proxies ever discovered, including failed ones
        self.__temp_proxies = []              # List of newly gathered Proxies, that still need to be checked
        self.__read_proxies()
        self.firstRun = True
        self.MIN_PROXIES_FOR_FIRST_RUN = 3

    def recheckProxies(self, forever=False):
        # Save firstRun locally because it will be changed in __verify_proxies:
        lFirst = self.firstRun
        self.__verify_proxies(self.__temp_proxies)
        if not self.reReadProxies:
            logger.info("recheckProxies was called, but reReadProxies is not set. " + \
                        "A bit of a strange combination of parameters.")
            return

        # If we're in first run and we found already the minimum amount of working proxies, return to calling
        # program
        if lFirst and len(self.proxies) >= self.MIN_PROXIES_FOR_FIRST_RUN:
            return

        self.__gather_proxies(self.__temp_proxies)
        if forever:
            t = Thread(target=self.__threaded_proxy)
            t.daemon = True
            logger.debug("Starting daemon process for threaded Proxy rechecks")
            t.start()
            logger.debug("Daemon process started")

    def __threaded_proxy(self):
        while True:
            if self.reReadProxies:
                self.__gather_proxies()
            else:
                msg = "This thread quits now, as reReadProxies is set to false - nothing to do. Most probably " + \
                      "shouldn't  have been a thread in the first place."
                logger.critical(msg)
                sys.exit(msg)

    def __gather_proxies(self, proxy=None, test=False):
        if proxy == None:
            self.__temp_proxies = [self.proxies[p] for p in self.proxies]
        else:
            self.__temp_proxies = proxy
        logger.debug("Checking for new proxies...")
        try:
            response = requests.get(self.proxy_gather_link, timeout=15)
        except Exception as ex:
            print(ex)
            logger.error(f"No proxies read. Maybe Internet-connection is down. Please check and retry. Error was {ex}")
            return None

        try:
            if response.status_code < 400:
                soup = bs(response.content, 'html.parser')
                table = soup.find('tbody')
                tr_list = table.find_all('tr')
                for tr in tr_list:
                    ip = tr.find_all('td')[0].text
                    port = tr.find_all('td')[1].text
                    proxy = proxy_data().from_dict({"ip": ip, "port": port})
                    # Take over only new proxies.
                    if proxy not in self.__temp_proxies:
                        logger.debug(f"Added {proxy} to be checked ")
                        self.__temp_proxies.append(proxy)
                    else:
                        logger.debug(f"{proxy} gathered was already known")
            else:
                logger.debug("Error code 400 or greater in proxy link request")

        except Exception as ex:
            logger.error(f"Unable to parse html response. Error was {ex}")
        if test:
            return self.__temp_proxies

        self.__verify_proxies(self.__temp_proxies)

    def __verify_proxies(self, proxy_lis, test=False):
        removable = []
        goodProxies = 0
        links = ["https://www.youtube.com/watch?v=FghBQsZJg4U", "https://gogs.earthsquad.global"]
        headers = {
            'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/80.0.3987.163 Safari/537.36'
        }
        for lineCount, proxi in enumerate(proxy_lis):
            responses = []
            if not self.firstRun:
                # When the firstRun is over and we're continuously re-reading and checking proxies we don't want to
                # consume too much bandwidth. The execution will run in a separate thread, but still 5 Seconds brea
                # sounds good.
                time.sleep(5)
            logger.debug(f"{(str(proxy_lis.index(proxi) + 1))}/ {str(len(proxy_lis))}: {proxi}")
            proxy = self.__set_proxy(proxi)
            for link in links:
                res = self.__get_response(link, proxy, headers)
                responses.append(res)
            if False in responses:
                logger.debug(f"Error code 400 or greater in {proxi}")
                if test:
                    return f"Error code 400 or greater in {proxi}"
                if proxi.ip in self.all_proxies:
                    self.remove_proxy(proxi.ip)
                if proxi in self.__temp_proxies:
                    self.__temp_proxies.remove(proxi)
                continue

            try:
                soup1 = bs(responses[0].content, 'html.parser')
                soup2 = bs(responses[1].content, 'html.parser')
                name = soup1.find('meta', {'property': 'og:title'}).get('content')
                logo = soup2.find('div', class_='logo')
                if name == None or logo == None:
                    raise ValueError('Unable to parse html response.')

                logger.debug(f"Proxy can be used for Youtube: {proxi}")
                if test:
                    return f"Proxy can be used for Youtube: {proxi}"
                if proxi.ip not in self.proxies:
                    self.proxies[proxi.ip] = proxi
                    if proxi.ip not in self.all_proxies:
                        self.all_proxies[proxi.ip] = proxi
                elif proxi.failed <= GC.PROXY_FAILCOUNTER:
                    self.proxies[proxi.ip] = proxi
                # Check, if we have at least 4 proxies in the first run of the function.
                # If yes, exit, as the function will re-run soon.
                goodProxies += 1
                if goodProxies >= self.MIN_PROXIES_FOR_FIRST_RUN and self.firstRun:
                    self.firstRun = False
                    break
            except Exception as ex:
                logger.debug(f"Proxy not usable for Youtube: {proxi}, Exception: {ex}")
                if test:
                    return f"Proxy not usable for Youtube: {proxi}, Exception: {ex}"
                if proxi.ip in self.all_proxies:
                    self.remove_proxy(proxi.ip)
                if proxi in self.__temp_proxies:
                    self.__temp_proxies.remove(proxi)
            self.__write_proxies()
        logger.info(f"Total proxies in list currently = {str(len(self.all_proxies))}")
        logger.info(f"Identified {len(self.proxies)} working proxies")
        self.__write_proxies()

    def __get_response(self, link, proxy, headers):
        try:
            response = requests.get(link, proxies=proxy, headers=headers, timeout=20)
        except Exception as ex:
            logger.debug(f"Proxy not reacting: {proxy}")
            return False
        try:
            if response.status_code >= 400:
                return False
        except:
            return False
        return response

    def __set_proxy(self, proxi):
        if proxi.type == "socks5":
            if proxi.username != "" and proxi.password != "":
                proxy = {
                    "http": f"socks5://{proxi.username}:{proxi.password}@{proxi.ip}:{proxi.port}",
                    "https": f"socks5://{proxi.username}:{proxi.password}@{proxi.ip}:{proxi.port}",
                }
            else:
                proxy = {
                    "http": f"socks5://{proxi.ip}:{proxi.port}",
                    "https": f"socks5://{proxi.ip}:{proxi.port}",
                }
        elif proxi.username != "" and proxi.password != "":
            proxy = {
                "http": f"http://{proxi.username}:{proxi.password}@{proxi.ip}:{proxi.port}",
                "https": f"https://{proxi.username}:{proxi.password}@{proxi.ip}:{proxi.port}",
            }
        else:
            proxy = {
                "http": f"http://{proxi.ip}:{proxi.port}",
                "https": f"http://{proxi.ip}:{proxi.port}",
            }
        return proxy

    def __read_proxies(self):
        try:
            with open('proxies.csv','r') as csv_file:
                raw_data = csv.DictReader(csv_file)
                for data in raw_data:
                    # Take over all proxies from the file:
                    proxy = proxy_data().from_dict(dict(data))
                    # Take over only proxies, that used to work into __temp_proxies:
                    if int(data['failed']) < GC.PROXY_FAILCOUNTER:
                        self.__temp_proxies.append(proxy)
                    self.all_proxies[data["ip"]] = proxy
        except Exception as ex:
            logger.debug(str(ex))

    def __write_proxies(self):
        if len(self.all_proxies) > 0:
            with open("proxies.csv", 'w', newline='\n')as file:
                fl = csv.DictWriter(file, self.all_proxies[list(self.all_proxies.keys())[0]].to_dict().keys())
                fl.writeheader()
                for proxy in self.all_proxies:
                    fl.writerow(self.all_proxies[proxy].to_dict())
            logger.debug(f"Wrote list of {len(self.proxies)} proxies to CSV-File")

    def __getProxy(self):
        lMaxCount = 600
        lCount = 0
        if len(self.proxies) == 0:
            logger.info("Waiting for a working proxy.")
        while len(self.proxies) == 0 and lCount <= lMaxCount:
            lCount += 1
            time.sleep(1)
        logger.critical(f"Proxies count: {len(self.proxies)}")
        if len(self.proxies) == 0:
            logger.critical("No proxy was found. Maybe internet down?")
            raise BaseException("No proxy was found. Maybe internet down?")
        proxy = self.proxies[list(self.proxies.keys())[randint(0, len(self.proxies) - 1)]]
        proxy.Called()
        if proxy.username == "" and proxy.password == "":
            return {"ip": proxy.ip, "port": proxy.port, "type": proxy.type}
        else:
            return {
                "ip": proxy.ip, "port": proxy.port,
                "username": proxy.username, "password": proxy.password, "type": proxy.type
            }

    def random_proxy(self):
        return self.__getProxy()

    def remove_proxy(self, ip, port=None, type=None):
        try:
            logger.debug(f"Increase fail count on Proxy with type {type}: {ip}:{port}")
            self.all_proxies[ip].Failed()
            if self.all_proxies[ip].failed >= GC.PROXY_FAILCOUNTER:
                del self.proxies[ip]
                logger.debug(f"Ip {ip} removed successfully.")
        except Exception as ex:
            logger.error(str(ex))


    def testProxy(self,type, ip, port, user, password):
        return "Method not yet implemented"

    """
    # TODO Functions needed for integration test?
    def testGatherProxy(self):
        proxies = self.__gather_proxies(test=True)
        print(f"Total gathered untested proxies = {str(len(proxies))}")
        return proxies

    def testVerifyProxy(self):
        proxies_lis = []
        if len(self.__temp_proxies) > 0:
            proxies_lis = [x for x in self.__temp_proxies if not x.failed > GC.PROXY_FAILCOUNTER]
        if len(proxies_lis) < 1:
            proxies_lis = [self.testGatherProxy()]
        proxies_lis = [proxies_lis[randint(0, len(proxies_lis)-1)]]
        result = self.__verify_proxies(proxies_lis, test=True)
        return result
    """




try:
    import schwifty
except Exception as ex:
    pass

from baangt.base.Faker import Faker
import random


class IBAN:
    def __init__(self, bankLeitZahl='20151', bankLand='AT'):
        """
        Class to generate random IBAN-Numbers. Mainly used in test data generation either on-the-fly or upfront
        during creation of a testdata XLSX.

        @param bankLeitZahl: Valid Bankcode (old format)
        @param bankLand: Valid Bank country in ISO-Format
        """
        self.bankLeitZahl = bankLeitZahl
        self.bankLand = bankLand

    def getRandomIBAN(self):
        """
        Generates a random IBAN based on bankLand and bankLeitzahl as well as a random account number

        @return: gives a String of IBAN
        """
        laenge = random.randrange(6, 10)
        digits = []
        for n in range(laenge):
            digits.append(random.randrange(0, 10))
        digits = "".join(str(x) for x in digits)
        # Schwifty doesn't work on Ubuntu
        try:
            lReturn = str(schwifty.IBAN.generate(country_code=self.bankLand,
                                              bank_code=self.bankLeitZahl,
                                              account_code=digits))
        except Exception:
            lReturn = Faker().fakerProxyIBAN("AT")

        return lReturn

if __name__ == '__main__':
    l = IBAN()
    print(l.getRandomIBAN())



import baangt
from baangt.base.ExportResults.ExportResults import\
    (ExportResults, ExportTiming, ExcelSheetHelperFunctions)
import logging

logger = logging.getLogger("pyC")


class ExportResultsHookImpl:
    @baangt.hook_impl
    def exportResults_init(self, **kwargs):
        return ExportResults(**kwargs)

    @baangt.hook_impl
    def exportResults_exportToDataBase(self, exportResultsObject):
        return exportResultsObject.exportToDataBase()

    @baangt.hook_impl
    def exportResults_exportResult(self, exportResultsObject, **kwargs):
        return exportResultsObject.exportResultExcel(**kwargs)

    @baangt.hook_impl
    def exportResults_makeSummary(self, exportResultsObject):
        return exportResultsObject.makeSummaryExcel()

    @baangt.hook_impl
    def exportResults_closeExcel(self, exportResultsObject, line, cellNumber, testRecordDict, fieldName, strip=False):
        return exportResultsObject.closeExcel(line, cellNumber, testRecordDict, fieldName, strip)


class ExcelSheetHelperFunctionsHookImpl:
    
    @baangt.hook_impl
    def excelSheetHelperFunctions_init(self):
        return ExcelSheetHelperFunctions()

    @baangt.hook_impl
    def excelSheetHelperFunctions_set_column_autowidth(self, excelSheetHelperFunctionsObject, worksheet, column):
        return excelSheetHelperFunctionsObject.set_column_autowidth(worksheet, column)

    @baangt.hook_impl
    def excelSheetHelperFunctions_get_column_width(self, excelSheetHelperFunctionsObject, worksheet, column):
        return excelSheetHelperFunctionsObject.get_column_width(worksheet, column)


class ExportTimingHookImpl:
    @baangt.hook_impl
    def exportTiming_init(self, testdataRecords, sheet):
        return ExportTiming(testdataRecords, sheet)

    @baangt.hook_impl
    def exportTiming_writeHeader(self, exportTimingObject):
        return exportTimingObject.writeHeader()

    @baangt.hook_impl
    def exportTiming_writeLines(self, exportTimingObject):
        return exportTimingObject.writeLines()

    @baangt.hook_impl
    def exportTiming_shortenTimingValue(self, exportTimingObject, timingValue):
        return exportTimingObject.shortenTimingValue(timingValue)

    @baangt.hook_impl
    def exportTiming_writeCell(self, exportTimingObject, row, col, content, format=None):
        return exportTimingObject.writeCell(row, col, content, format)

    @baangt.hook_impl
    def exportTiming_findAllTimingSections(self, exportTimingObject):
        return exportTimingObject.findAllTimingSections()

    @baangt.hook_impl
    def exportTiming_interpretTimeLog(self, exportTimingObject, lTimeLog):
        return exportTimingObject.interpretTimeLog(lTimeLog)



import os
import logging
from time import sleep
from pathlib import Path
from datetime import datetime
from baangt.base.PathManagement import ManagedPaths
from baangt.base.DownloadFolderMonitoring import DownloadFolderMonitoring

logger = logging.getLogger("pyC")


class Cleanup:
    def __init__(self, days: float = 31):
        self.managedPaths = ManagedPaths()
        self.threshold = days

    def clean_logs(self):
        logger.info(f"Removing logs older than {str(self.threshold)} days")
        logs_dir = self.managedPaths.getLogfilePath()
        files = Path(logs_dir).glob('**/*')
        removed = []
        for file in files:
            timedelta = datetime.now() - datetime.fromtimestamp(DownloadFolderMonitoring.creation_date(file))
            if timedelta.total_seconds()/86400 > self.threshold:
                try:
                    os.remove(str(file))
                    removed.append(file)
                except:
                    logger.debug(f"Cannot remove {str(file)}")
                    continue
                logger.debug(f"{str(file)} deleted")
        if len(removed) == 0:
            logger.info(f"No log file older than {str(self.threshold)} days found")
        else:
            logger.info(f"{str(len(removed))} Log file deleted")

    def clean_screenshots(self):
        logger.info(f"Removing screenshots older than {str(self.threshold)} days")
        ss_dir = self.managedPaths.getOrSetScreenshotsPath()
        files = Path(ss_dir).glob('**/*')
        removed = []
        for file in files:
            timedelta = datetime.now() - datetime.fromtimestamp(DownloadFolderMonitoring.creation_date(file))
            if timedelta.total_seconds()/86400 > self.threshold:
                try:
                    os.remove(str(file))
                    removed.append(file)
                except:
                    logger.debug(f"Cannot remove {str(file)}")
                    continue
                logger.debug(f"{str(file)} deleted")
        if len(removed) == 0:
            logger.info(f"No Screenshot older than {str(self.threshold)} days found")
        else:
            logger.info(f"{str(len(removed))} Screenshot deleted")

    def clean_downloads(self):
        logger.info(f"Removing downloads older than {str(self.threshold)} days")
        downloads_dir = self.managedPaths.getOrSetAttachmentDownloadPath()
        files = Path(downloads_dir).glob('**/*')
        removed = []
        for file in files:
            if file.is_file():
                timedelta = datetime.now() - datetime.fromtimestamp(DownloadFolderMonitoring.creation_date(file))
                if timedelta.total_seconds()/86400 > self.threshold:
                    try:
                        os.remove(str(file))
                        removed.append(file)
                    except:
                        logger.debug(f"Cannot remove {str(file)}")
                        continue
                    logger.debug(f"{str(file)} deleted")
        if len(removed) == 0:
            logger.info(f"No downloads older than {str(self.threshold)} days found")
        else:
            logger.info(f"{str(len(removed))} downloads deleted")
        sleep(1)
        logger.info("Removing empty folders inside download folder.")
        files = Path(downloads_dir).glob('**/*')
        removed = []
        for file in files:
            if file.is_dir():
                if len(os.listdir(str(file))) == 0:
                    try:
                        os.rmdir(str(file))
                        removed.append(file)
                    except:
                        logger.debug(f"Cannot remove folder {str(file)}")
                        continue
                    logger.debug(f"{str(file)} folder deleted")
        if len(removed) == 0:
            logger.info(f"No empty folder found inside downloads")
        else:
            logger.info(f"{str(len(removed))} empty folder deleted")

    def clean_all(self):
        self.clean_logs()
        self.clean_screenshots()
        self.clean_downloads()



import baangt.base.GlobalConstants as GC
from baangt.base.Timing.Timing import Timing
from baangt.base.BrowserHandling.BrowserHandling import BrowserDriver
from baangt.base.ApiHandling import ApiHandling
import sys
from pkg_resources import parse_version
import logging
from baangt.TestSteps.Exceptions import baangtTestStepException
from baangt.TestSteps.AddressCreation import AddressCreate
from baangt.base.PDFCompare import PDFCompare, PDFCompareDetails
from baangt.base.Faker import Faker as baangtFaker
from baangt.base.Utils import utils
from baangt.base.RuntimeStatistics import Statistic
import random
import itertools

logger = logging.getLogger("pyC")


class TestStepMaster:
    def __init__(self, executeDirect=True, **kwargs):
        self.anchor = None
        self.anchorLocator = None
        self.anchorLocatorType = None
        self.testCaseStatus = None
        self.ifActive = False
        self.ifIsTrue = True
        self.elseIsTrue = False
        self.ifLis = [self.ifIsTrue]
        self.elseLis = [self.elseIsTrue]
        self.ifConditions = 0
        self.repeatIsTrue = [False]
        self.repeatDict = [] # used to store steps command of repeat data
        self.repeatData = [] # used to store RLP_ data to be looped in repeat
        self.repeatCount = [] # Used to store count of randomdata in loop
        self.repeatActive = 0 # to sync active repeat counts with repeat done and will be execute when both are equal
        self.repeatDone = 0
        self.baangtFaker = None
        self.statistics = Statistic()
        self.kwargs = kwargs
        self.testRunInstance = kwargs.get(GC.KWARGS_TESTRUNINSTANCE)
        self.testcaseDataDict = kwargs.get(GC.KWARGS_DATA)
        self.timing: Timing = kwargs.get(GC.KWARGS_TIMING)
        self.timingName = self.timing.takeTime(self.__class__.__name__, forceNew=True)
        self.browserSession: BrowserDriver = kwargs.get(GC.KWARGS_BROWSER)
        self.apiSession: ApiHandling = kwargs.get(GC.KWARGS_API_SESSION)

        self.testStepNumber = kwargs.get(GC.STRUCTURE_TESTSTEP)  # Set in TestRunData by TestCaseMaster
        self.testRunUtil = self.testRunInstance.testRunUtils

        # Get the release from testrun-customizing if set: (will influence, whether or not elements will be executed)
        self.globalRelease = self.testRunInstance.globalSettings.get("Release", "")

        # check, if this TestStep has additional Parameters and if so, execute
        lSequence = self.testRunUtil.getSequenceByNumber(testRunName=self.testRunInstance.testRunName,
                                                         sequence=kwargs.get(GC.STRUCTURE_TESTCASESEQUENCE))
        self.testCase = self.testRunUtil.getTestCaseByNumber(lSequence, kwargs.get(GC.STRUCTURE_TESTCASE))
        self.testStep = self.testRunUtil.getTestStepByNumber(self.testCase, self.testStepNumber)

        if self.testStep and len(self.testStep) > 1:
            if not isinstance(self.testStep[1], str) and executeDirect:
                # This TestStepMaster-Instance should actually do something - activitites are described
                # in the TestExecutionSteps.
                # Otherwise there's only a classname in TestStep[0]
                self.executeDirect(self.testStep[1][GC.STRUCTURE_TESTSTEPEXECUTION])

                # Teardown makes only sense, when we actually executed something directly in here
                # Otherwise (if it was 1 or 2 Tab-stops more to the left) we'd take execution time without
                # having done anything
                self.teardown()

        self.statistics.update_teststep_sequence()

    def executeDirect(self, executionCommands):
        """
        Executes a sequence of Commands. Will be subclassed in other modules.
        :param executionCommands:
        :return:
        """
        for index, (key, command) in enumerate(executionCommands.items()):
            try:
                self.executeDirectSingle(index, command)
            except Exception as ex:
                # 2020-07-16: This Exception is cought, printed and then nothing. That's not good. The test run
                # continues forever, despite this exception. Correct way would be to set test case to error and stop
                # this test case
                # Before we change something here we should check, why the calling function raises an error.
                logger.critical(ex)
                self.testcaseDataDict[GC.TESTCASESTATUS] = GC.TESTCASESTATUS_ERROR
                return

    def manageNestedCondition(self, condition="", ifis=False):
        if condition.upper() == "IF":
            self.ifConditions += 1
            self.ifLis.append(ifis)
            self.elseLis.append(False)
        elif condition.upper() == "ELSE":
            self.elseLis[-1] = True
        elif condition.upper() == "ENDIF":
            self.ifConditions -= 1
            if self.ifConditions < 0:
                raise BaseException("Numbers of ENDIF are greater than IF.")
            self.ifLis.pop()
            self.elseLis.pop()
        assert len(self.ifLis) == self.ifConditions + 1 and len(self.elseLis) == self.ifConditions + 1
        self.ifIsTrue = self.ifLis[-1]
        self.elseIsTrue = self.elseLis[-1]

    def executeDirectSingle(self, commandNumber, command, replaceFromDict=None):
        """
        This will execute a single instruction
        """

        # when we have an IF-condition and it's condition was not TRUE, then skip whatever comes here until we
        # reach Endif
        if not self.ifIsTrue and not self.elseIsTrue:
            if command["Activity"].upper() == "IF":
                self.manageNestedCondition(condition=command["Activity"].upper(), ifis=self.ifIsTrue)
                return True
            if command["Activity"].upper() != "ELSE" and command["Activity"].upper() != "ENDIF":
                return True
        if self.repeatIsTrue[-1]: # If repeat statement is active then execute this
            if command["Activity"].upper() == "REPEAT": # To sync active repeat with repeat done
                self.repeatActive += 1
            if command["Activity"].upper() != "REPEAT-DONE": # store command in repeatDict
                self.repeatDict[-1][commandNumber] = command
                return
            else:
                self.repeatDone += 1 # to sync repeat done with active repeat
                if self.repeatDone < self.repeatActive: # if all repeat-done are not synced with repeat store the data
                    self.repeatDict[-1][commandNumber] = command
                    logger.info(command)
                    return
                self.repeatIsTrue[-1] = False
                if self.repeatData[-1]:
                    data_list = []
                    if type(self.repeatData[-1]) is list:
                        for data_dic in self.repeatData[-1]:
                            keys, values = zip(*data_dic.items())
                            final_values = []
                            for value in values: # coverting none list values to list. Useful in make all possible data using itertools
                                if type(value) is not list:
                                    final_values.append([value])
                                else:
                                    final_values.append(value)
                            data_l = [dict(zip(keys, v)) for v in itertools.product(*final_values)] # itertools to make all possible combinations
                            data_list.extend(data_l)
                    else:
                        data_list = [self.repeatData[-1]]
                    if len(self.repeatCount) > 0 and self.repeatCount[-1]: # get random data from list of data
                        try:
                            data_list = random.sample(data_list, int(self.repeatCount[-1]))
                        except:
                            pass
                    if data_list:
                        for data in data_list:
                            temp_dic = dict(self.repeatDict[-1])
                            processed_data = {}
                            for key in temp_dic:
                                try:
                                    processed_data = dict(temp_dic[key])
                                except Exception as ex:
                                    logger.debug(ex)
                                try:
                                    self.executeDirectSingle(key, processed_data, replaceFromDict=data)
                                except Exception as ex:
                                    logger.info(ex)
                    else:
                        temp_dic = dict(self.repeatDict[-1])
                        for key in temp_dic:
                            try:
                                processed_data = dict(temp_dic[key])
                            except Exception as ex:
                                logger.debug(ex)
                            for ky in processed_data:
                                try:
                                    processed_data[ky] = self.replaceVariables(processed_data[ky])
                                except Exception as ex:
                                    logger.info(ex)
                            try:
                                self.executeDirectSingle(key, processed_data)
                            except Exception as ex:
                                logger.info(ex)
                del self.repeatIsTrue[-1]
                del self.repeatDict[-1]
                del self.repeatData[-1]
                del self.repeatCount[-1]
                self.repeatDone -= 1
                self.repeatActive -= 1
                return

        css, id, lActivity, lLocator, lLocatorType, xpath = self._extractAllSingleValues(command)

        if lActivity == "COMMENT":
            return  # Comment's are ignored

        if self.anchor and xpath:
            if xpath[0:3] == '///':         # Xpath doesn't want to use Anchor
                xpath = xpath[1:]
                logger.debug(f"Anchor active, but escaped. Using pure xpath: {xpath}")
            else:
                logger.debug(f"Anchor active. combining {self.anchorLocator} with {xpath}")
                xpath = self.anchorLocator + xpath

        lValue = str(command["Value"])
        lValue2 = str(command["Value2"])
        lComparison = command["Comparison"]
        lOptional = utils.anything2Boolean(command["Optional"])

        # check release line
        lRelease = command["Release"]

        # Timeout defaults to 20 seconds, if not set otherwise.
        lTimeout = TestStepMaster._setTimeout(command["Timeout"])

        lTimingString = f"TS {commandNumber} {lActivity.lower()}"
        self.timing.takeTime(lTimingString)
        logger.info(
            f"Executing TestStepDetail {commandNumber} with parameters: act={lActivity}, lType={lLocatorType}, loc={lLocator}, "
            f"Val1={lValue}, comp={lComparison}, Val2={lValue2}, Optional={lOptional}, timeout={lTimeout}")
        original_value = lValue # used as key to make rlp json dict and will be used further to make sheet name
        lValue, lValue2 = self.replaceAllVariables(lValue, lValue2, replaceFromDict=replaceFromDict)

        if not TestStepMaster.ifQualifyForExecution(self.globalRelease, lRelease):
            logger.debug(f"we skipped this line due to {lRelease} disqualifies according to {self.globalRelease} ")
            return
        if lActivity == "GOTOURL":
            self.browserSession.goToUrl(lValue)
        elif lActivity == "SETTEXT":
            self.browserSession.findByAndSetText(xpath=xpath, css=css, id=id, value=lValue, timeout=lTimeout)
        elif lActivity == "SETTEXTIF":
            self.browserSession.findByAndSetTextIf(xpath=xpath, css=css, id=id, value=lValue, timeout=lTimeout,
                                                   optional=lOptional)
        elif lActivity == "FORCETEXT":
            self.browserSession.findByAndForceText(xpath=xpath, css=css, id=id, value=lValue, timeout=lTimeout,
                                                   optional=lOptional)
        elif lActivity == "FORCETEXTIF":
            if lValue:
                self.browserSession.findByAndForceText(xpath=xpath, css=css, id=id, value=lValue, timeout=lTimeout,
                                                       optional=lOptional)
        elif lActivity == "SETANCHOR":
            if not lLocator:
                self.anchor = None
                self.anchorLocator = None
                self.anchorLocatorType = None
            else:
                found = self.browserSession.findBy(xpath=xpath, css=css, id=id, timeout=lTimeout)
                if found:
                    self.anchor = self.browserSession.element
                    self.anchorLocator = lLocator
                    self.anchorLocatorType = lLocatorType
                    logger.debug(f"Anchor set: {lLocatorType} {lLocator}")
                else:
                    logger.error(f"Anchor should be set, but can't be found in the current page: {lLocatorType}, {lLocator}")
                    raise ValueError(f"Anchor should be set, but can't be found in the current page: {lLocatorType}, {lLocator}")
        elif lActivity == 'HANDLEIFRAME':
            self.browserSession.handleIframe(lLocator)
        elif lActivity == 'SWITCHWINDOW':
            lWindow = lValue
            if lWindow.isnumeric():
                lWindow = int(lWindow)
            self.browserSession.handleWindow(windowNumber=lWindow, timeout=lTimeout)
        elif lActivity == "CLICK":
            self.browserSession.findByAndClick(xpath=xpath, css=css, id=id, timeout=lTimeout)
        elif lActivity == "CLICKIF":
            self.browserSession.findByAndClickIf(xpath=xpath, css=css, id=id, timeout=lTimeout, value=lValue)
        elif lActivity == "PAUSE":
            self.browserSession.sleep(seconds=float(lValue))
        elif lActivity == "IF":
            # Originally we had only Comparisons. Now we also want to check for existance of Field
            if not lValue and lLocatorType and lLocator:
                lValue = self.browserSession.findBy(xpath=xpath, css=css, id=id, optional=lOptional,
                                                    timeout=lTimeout)

            self.__doComparisons(lComparison=lComparison, value1=lValue, value2=lValue2)
            logger.debug(f"IF-condition {lValue} {lComparison} {lValue2} evaluated to: {self.ifIsTrue} ")
        elif lActivity == "ELSE":
            if not self.ifIsTrue:
                self.manageNestedCondition(condition=lActivity)
                logger.debug("Executing ELSE-condition")
        elif lActivity == "ENDIF":
            self.manageNestedCondition(condition=lActivity)
        elif lActivity == "REPEAT":
            self.repeatActive += 1
            self.repeatIsTrue.append(True)
            self.repeatDict.append({})
            if original_value not in self.testRunInstance.json_dict:
                self.testRunInstance.json_dict[original_value] = []
            self.testRunInstance.json_dict[original_value].append(lValue)
            self.repeatData.append(lValue)
            self.repeatCount.append(lValue2)
        elif lActivity == 'GOBACK':
            self.browserSession.goBack()
        elif lActivity == 'APIURL':
            self.apiSession.setBaseURL(lValue)
        elif lActivity == 'ENDPOINT':
            self.apiSession.setEndPoint(lValue)
        elif lActivity == 'POST':
            self.apiSession.postURL(content=lValue)
        elif lActivity == 'GET':
            self.apiSession.getURL()
        elif lActivity == 'HEADER':
            self.apiSession.setHeaders(setHeaderData=lValue)
        elif lActivity == 'SAVE':
            self.doSaveData(lValue, lValue2, lLocatorType, lLocator)
        elif lActivity == 'CLEAR':
            # Clear a variable:
            if self.testcaseDataDict.get(lValue):
                del self.testcaseDataDict[lValue]
        elif lActivity == 'SAVETO':
            # In this case, we need to parse the real field, not the representation of the replaced field value
            self.doSaveData(command['Value'], lValue2, lLocatorType, lLocator)
        elif lActivity == 'SUBMIT':
            self.browserSession.submit()
        elif lActivity == "ADDRESS_CREATE":
            # Create Address with option lValue and lValue2
            AddressCreate(lValue, lValue2)
            # Update testcaseDataDict with addressDict returned from
            AddressCreate.returnAddress()
            self.testcaseDataDict.update(AddressCreate.returnAddress())
        elif lActivity == 'ASSERT':
            value_found = self.browserSession.findByAndWaitForValue(xpath=xpath, css=css, id=id, optional=lOptional,
                                                                    timeout=lTimeout)
            if not self.__doComparisons(lComparison=lComparison, value1=value_found, value2=lValue):
                raise baangtTestStepException(f"Expected Value: {lValue}, Value found :{value_found} ")
        elif lActivity == 'IBAN':
            # Create Random IBAN. Value1 = Input-Parameter for IBAN-Function. Value2=Fieldname
            self.__getIBAN(lValue, lValue2)
        elif lActivity == 'PDFCOMPARE':
            self.doPDFComparison(lValue)
        elif lActivity == 'CHECKLINKS':
            self.checkLinks()
        elif lActivity == 'ALERTIF':
            self.browserSession.confirmAlertIfAny()
        elif lActivity == GC.TESTCASESTATUS_STOP.upper():
            self.testcaseDataDict[GC.TESTCASESTATUS_STOP] = "X"              # will stop the test case
        elif lActivity == GC.TESTCASESTATUS_STOPERROR.upper():
            self.testcaseDataDict[GC.TESTCASESTATUS_STOPERROR] = "X"         # will stop the test case and set error
        elif lActivity[0:3] == "ZZ_":
            # custom command. Do nothing and return
            return
        else:
            raise BaseException(f"Unknown command in TestStep {lActivity}")

        self.timing.takeTime(lTimingString)

    def _extractAllSingleValues(self, command):
        lActivity = command["Activity"].upper()
        lLocatorType = command["LocatorType"].upper()
        try:
            lLocator = self.replaceVariables(command["Locator"])
        except Exception as ex:
            logger.info(ex)
        if lLocator and not lLocatorType:  # If locatorType is empty, default it to XPATH
            lLocatorType = 'XPATH'
        xpath, css, id = self.__setLocator(lLocatorType, lLocator)
        return css, id, lActivity, lLocator, lLocatorType, xpath

    def doPDFComparison(self, lValue, lFieldnameForResults="DOC_Compare"):
        lFiles = self.browserSession.findNewFiles()
        if len(lFiles) > 1:
            # fixme: Do something! There were more than 1 files since last check. Damn
            logger.critical(f"There were {len(lFiles)} files new since last check. Can't handle that. ")
            raise Exception
        elif len(lFiles) == 1:
            # Wonderful. Let's do the PDF-Comparison
            lPDFDataClass = PDFCompareDetails()
            lPDFDataClass.fileName = lFiles[0][0]
            lPDFDataClass.referenceID = lValue
            lDict = {"": lPDFDataClass}
            lPDFCompare = PDFCompare()
            lDict = lPDFCompare.compare_multiple(lDict)
            self.testcaseDataDict[lFieldnameForResults + "_Status"] = lDict[""].Status
            self.testcaseDataDict[lFieldnameForResults + "_Results"] = lDict[""].StatusText

    def replaceAllVariables(self, lValue, lValue2, replaceFromDict=None):
        # Replace variables from data file
        try:
            if len(lValue) > 0:
                lValue = self.replaceVariables(lValue, replaceFromDict=replaceFromDict)
            if len(lValue2) > 0:
                lValue2 = self.replaceVariables(lValue2, replaceFromDict=replaceFromDict)
        except Exception as ex:
            logger.info(ex)
        return lValue, lValue2

    def __getIBAN(self, lValue, lValue2):
        from baangt.base.IBAN import IBAN
        if not lValue2:
            logger.critical("IBAN-Method called without destination field name in column 'Value 2'")
            return

        lIBAN = IBAN()
        self.testcaseDataDict[lValue2] = lIBAN.getRandomIBAN()
        pass

    def checkLinks(self):
        """
        Will check all links on the current webpage

        Result will be written into "CheckedLinks" in TestDataDict

        If theres a returncode >= 400 in the list, we'll mark the testcase as failed
        """
        if self.testcaseDataDict.get("CheckedLinks"):
            self.testcaseDataDict["Result_CheckedLinks"] += self.browserSession.checkLinks()
        else:
            self.testcaseDataDict["Result_CheckedLinks"] = self.browserSession.checkLinks()

        for entry in self.testcaseDataDict["Result_CheckedLinks"]:
            if entry[0] > 400:
                self.testcaseDataDict[GC.TESTCASESTATUS] = GC.TESTCASESTATUS_ERROR
                self.testcaseDataDict[GC.TESTCASEERRORLOG] = self.testcaseDataDict.get(GC.TESTCASEERRORLOG, "") \
                                                             + "URL-Checker error"
                break

    @staticmethod
    def ifQualifyForExecution(version_global, version_line):
        """ This function will test version_global and version_line
            @return True or False
        """
        if not version_global.strip():
            # No value is defined in Release, return True
            return True
        if not version_line.strip():
            # we skipped this line
            return True

        # split the version line
        if not len(version_line.strip().split(" ")) == 2:
            logger.debug(f"Invalid release format {version_line} ")
            return True
        comp_operator, version = version_line.strip().split(" ")
        if comp_operator == "<":
            return parse_version(version_global) < parse_version(version)
        elif comp_operator == ">":
            return parse_version(version_global) > parse_version(version)
        elif comp_operator == ">=":
            return parse_version(version_global) >= parse_version(version)
        elif comp_operator == "<=":
            return parse_version(version_global) <= parse_version(version)
        elif comp_operator == "=" or comp_operator == "==":
            return parse_version(version_global) == parse_version(version)
        else:
            logger.debug(f"Global version {version_global}, line version {version_line} ")
            return False

    def doSaveData(self, toField, valueForField, lLocatorType, lLocator):
        """
        Save fields. Either from an existing DICT (usually in API-Mode) or from a Webelement (in Browser-Mode)

        :param toField:
        :param valueForField:
        :param lLocatorType:
        :param lLocator:
        :return: no return parameter. The implicit return is a value in a field.
        """
        if self.testCase[1][GC.KWARGS_TESTCASETYPE] == GC.KWARGS_BROWSER:
            xpath, css, id = TestStepMaster.__setLocator(lLocatorType, lLocator)
            self.testcaseDataDict[toField] = self.browserSession.findByAndWaitForValue(xpath=xpath, css=css, id=id)
            logger.debug(f"Saved {self.testcaseDataDict[toField]} to {toField}")
        elif self.testCase[1][GC.KWARGS_TESTCASETYPE] == GC.KWARGS_API_SESSION:
            self.testcaseDataDict[toField] = valueForField
        else:
            sys.exit("Testcase Type not supported")

    @staticmethod
    def __setLocator(lLocatorType, lLocator):
        return utils.setLocatorFromLocatorType(lLocatorType, lLocator)

    @staticmethod
    def _setTimeout(lTimeout):
        return 20 if not lTimeout else float(lTimeout)

    def __doComparisons(self, lComparison, value1, value2):
        if isinstance(value1, bool) or isinstance(value2, bool):
            value1 = utils.anything2Boolean(value1)
            value2 = utils.anything2Boolean(value2)

        if value2 == 'None':
            value2 = None

        if lComparison == "=":
            if value1 == value2:
                self.manageNestedCondition(condition="IF", ifis=True)
            else:
                self.manageNestedCondition(condition="IF", ifis=False)
        elif lComparison == "!=":
            self.ifIsTrue = False if value1 == value2 else True
        elif lComparison == ">":
            if value1 > value2:
                self.manageNestedCondition(condition="IF", ifis=True)
            else:
                self.manageNestedCondition(condition="IF", ifis=False)
        elif lComparison == "<":
            if value1 < value2:
                self.manageNestedCondition(condition="IF", ifis=True)
            else:
                self.manageNestedCondition(condition="IF", ifis=False)
        elif not lComparison:  # Check only, if Value1 has a value.
            val = True if value1 else False
            self.manageNestedCondition(condition="IF", ifis=val)
        else:
            raise BaseException(f"Comparison Operator not supported/unknown {lComparison}")

        return self.ifIsTrue

    def execute(self):
        """Method is overwritten in all children/subclasses"""
        pass

    def teardown(self):
        if self.testCaseStatus:
            if not self.testcaseDataDict[GC.TESTCASESTATUS]:
                self.testcaseDataDict[GC.TESTCASESTATUS] = self.testCaseStatus
            elif self.testCaseStatus == GC.TESTCASESTATUS_SUCCESS:
                # We'll not overwrite a potential Error Status of the testcase
                pass
            elif self.testCaseStatus == GC.TESTCASESTATUS_ERROR:
                self.testcaseDataDict = GC.TESTCASESTATUS_ERROR
            else:
                sys.exit("No idea, what happened here. Unknown condition appeared")

        self.timing.takeTime(self.timingName)

    def replaceVariables(self, expression, replaceFromDict=None):
        """
        The syntax for variables is currently $(<column_name_from_data_file>). Multiple variables can be assigned
        in one cell, for instance perfectly fine: "http://$(BASEURL)/$(ENDPOINT)"

        There's a special syntax for the faker module: $(FAKER.<fakermethod>).

        Also a special syntax for API-Handling: $(APIHandling.<DictElementName>).

        @param expression: the current cell, either as fixed value, e.g. "Franzi" or with a varible $(DATE)
        @return: the replaced value, e.g. if expression was $(DATE) and the value in column "DATE" of data-file was
            "01.01.2020" then return will be "01.01.2020"
        """
        if "$(" not in expression:
            return expression

        while "$(" in expression:
            if expression[0:2] == "$(":
                left_part = ""
            else:
                left_part = expression.split("$(")[0]

            center = expression[len(left_part) + 2:]
            center = center.split(")")[0]

            right_part = expression[len(left_part) + len(center) + 3:]
            centerValue = ""

            if replaceFromDict: # json is supplied with repeat tag, that json is used here to get main data
                    dic = replaceFromDict
                    for key in center.split('.')[-1:]:
                        dic = self.iterate_json(dic, key)
                    centerValue = dic

            if centerValue: # if we got value from the passed json then bypass this if else conditions
                pass
            elif "." not in center:
                # Replace the variable with the value from data structure
                centerValue = self.testcaseDataDict.get(center)
            else:
                # This is a reference to a DICT with ".": for instance APIHandling.AnswerContent("<bla>")
                dictVariable = center.split(".")[0]
                dictValue = center.split(".")[1]

                if dictVariable == 'ANSWER_CONTENT':
                    centerValue = self.apiSession.session[1].answerJSON.get(dictValue, "Empty")
                elif dictVariable == 'FAKER':
                    # This is to call Faker Module with the Method, that is given after the .
                    centerValue = self._getFakerData(dictValue)
                elif self.testcaseDataDict.get(dictVariable):
                    dic = self.testcaseDataDict.get(dictVariable)
                    for key in center.split('.')[1:]:
                        dic = self.iterate_json(dic, key)
                    centerValue = dic
                else:
                    raise BaseException(f"Missing code to replace value for: {center}")

            if not centerValue:
                if center in self.testcaseDataDict.keys():
                    # The variable exists, but has no value.
                    centerValue = ""
                else:
                    raise BaseException(f"Variable not found: {center}, input parameter was: {expression}")

            expression = "".join([left_part, str(centerValue), right_part])

        return expression

    def iterate_json(self, data, key):
        # itereate through list of json and create list of data from dictionary inside those list
        if type(data) is list:
            lis = []
            for dt in data:
                dt = self.iterate_json(dt, key)
                lis.append(dt)
            return lis
        elif type(data) is dict:
            return data.get(key)


    def _getFakerData(self, fakerMethod):
        if not self.baangtFaker:
            self.baangtFaker = baangtFaker()

        logger.debug(f"Calling faker with method: {fakerMethod}")

        return self.baangtFaker.fakerProxy(fakerMethod=fakerMethod)



from logging import getLogger
import baangt.base.GlobalConstants as GC
from icopy2xls import Mover
from baangt.base.PathManagement import ManagedPaths

logger = getLogger("pyC")


class Append2BaseXLS:
    """
    If in the globals of the current testrun the parameter AR2BXLS (Append Results to Base XLS) is set,
    we execute baangt Move-Corresponding module accordingly.
    """
    def __init__(self, testRunInstance, resultsFileName: str=None):
        self.testRunInstance = testRunInstance
        self.resultsFileName = resultsFileName
        self.mp = ManagedPaths()

        self._append2BaseXLS()

    def _append2BaseXLS(self):
        lGlobals = self.testRunInstance.globalSettings
        if not lGlobals.get("AR2BXLS"):
            logger.debug("No request to save to further destinations. Exiting.")
            return

        # Format: fileAndPath,Sheet;fileAndPath,sheet
        fileTuples = []
        if ";" in lGlobals.get("AR2BXLS"):
            files = lGlobals.get("AR2BXLS").split(";")
            for file in files:
                fileTuples.append(self.checkAppend(file))
        else:
            fileTuples.append(self.checkAppend(lGlobals["AR2BXLS"]))

        for fileTuple in fileTuples:
            if not fileTuple:
                logger.critical("File to append results to not found (see message above")
                break
            logger.info(f"Starting to append results to: {str(fileTuple)}")
            lMover = Mover(source_file_path=self.resultsFileName,
                           source_sheet="Output",
                           destination_file_path=fileTuple[0],
                           destination_sheet=fileTuple[1].strip())
            lMover.move(filters={GC.TESTCASESTATUS:GC.TESTCASESTATUS_SUCCESS}, add_missing_columns=False)
            logger.debug(f"Appending results to {str(fileTuple)} finished")

    def checkAppend(self, file):
        lFileAndPath = self.mp.findFileInAnyPath(filename=file.split(",")[0])
        if lFileAndPath:
            return [lFileAndPath, file.split(",")[1]]
        else:
            logger.critical(f"File not found anywhere: {file.split(',')[0]}")
            return None


from baangt.base.BrowserHandling.BrowserHandling import BrowserDriver
from baangt.base.ApiHandling import ApiHandling
from baangt.base.ExportResults.ExportResults import ExportResults
from baangt.base.Utils import utils
from baangt.base import GlobalConstants as GC
from baangt.base.TestRunExcelImporter import TestRunExcelImporter
from baangt.base.BrowserFactory import BrowserFactory
# needed - they'll be used dynamically later
from baangt.TestSteps.TestStepMaster import TestStepMaster
from baangt.TestCase.TestCaseMaster import TestCaseMaster
from baangt.TestCaseSequence.TestCaseSequenceMaster import TestCaseSequenceMaster
from baangt.base.ProxyRotate import ProxyRotate
from baangt.base.FilesOpen import FilesOpen
import xlsxwriter
import os
import json
import logging
from pathlib import Path
import sys
from baangt.base.Timing.Timing import Timing
from baangt.base.TestRunUtils import TestRunUtils
from baangt.base.TestRun.ClassesForObjects import ClassesForObjects
import time
from baangt.base.PathManagement import ManagedPaths
from uuid import uuid4
from baangt.base.RuntimeStatistics import Statistic
from baangt.base.SendReports import Sender
import signal

logger = logging.getLogger("pyC")


class TestRun:
    """
    This is the main Class of Testexecution in the baangt Framework. It is usually started
    from baangtIA.py
    """

    def __init__(self, testRunName, globalSettingsFileNameAndPath=None,
                 testRunDict=None, uuid=uuid4(), executeDirect=True):  # -- API support: testRunDict --
        """
        @param testRunName: The name of the TestRun to be executed.
        @param globalSettingsFileNameAndPath: from where to read the <globals>.json
        """

        # Take over importing parameters:
        self.uuid = uuid
        logger.info(f'Init Testrun, uuid is {self.uuid}')
        self.testRunDict = testRunDict
        self.globalSettingsFileNameAndPath = globalSettingsFileNameAndPath
        self.testRunName, self.testRunFileName = \
            self._sanitizeTestRunNameAndFileName(testRunName)

        # Initialize everything else
        self.apiInstance = None
        self.testType = None
        self.networkInfo = None
        self.results = None
        self.browserFactory = None
        self.kwargs = {}
        self.dataRecords = {}
        self.globalSettings = {}
        self.json_dict = {} # Used to maintain records of RLP_ data which will be used will exporting results
        self.managedPaths = ManagedPaths()
        self.classesForObjects = ClassesForObjects()         # Dynamically loaded classes
        self.timing = Timing()
        self.testRunUtils = TestRunUtils()
        self.testCasesEndDateTimes_1D = []                   # refer to single execution
        self.testCasesEndDateTimes_2D = [[]]                 # refer to parallel execution
        # New way to export additional Tabs to Excel
        # If you want to export additional data, place a Dict with Tabname + Datafields in additionalExportTabs
        # from anywhere within your custom code base.
        self.additionalExportTabs = {}
        self.statistics = Statistic()
        signal.signal(signal.SIGINT, self.exit_signal_handler)
        signal.signal(signal.SIGTERM, self.exit_signal_handler)

        # Initialize other values
        self.timing.takeTime(GC.TIMING_TESTRUN)               # Initialize Testrun Duration

        # Usually the Testrun is called without the parameter executeDirect, meaning it default to "Execute"
        # during Unit-Tests we don't want this behaviour:
        if executeDirect:
            self.executeTestRun()

    def exit_signal_handler(self, signal, frame):
        self.browserFactory.teardown()

    def executeTestRun(self):
        self._initTestRunSettingsFromFile()  # Loads the globals*.json file

        self._loadJSONTestRunDefinitions()
        self._loadExcelTestRunDefinitions()

        self.browserFactory = BrowserFactory(self)
        self.executeTestSequence()
        self.tearDown()

        try:
            Sender.send_all(self.results, self.globalSettings)
        except Exception as ex:
            logger.error(f"Error from SendAll: {ex}")

    def append1DTestCaseEndDateTimes(self, dt):
        self.testCasesEndDateTimes_1D.append(dt)

    def append2DTestCaseEndDateTimes(self, index, tcAndDt):
        tc = tcAndDt[0]
        dt = tcAndDt[1]
        [self.testCasesEndDateTimes_2D.append([]) for i in range(
            index + 1 - len(self.testCasesEndDateTimes_2D))] if index + 1 > len(
            self.testCasesEndDateTimes_2D) else None
        self.testCasesEndDateTimes_2D[index].append([tc, dt])

    def tearDown(self):
        """
        Close browser (unless stated in the Globals to not do so) and API-Instances
        Take overall Time spent for the complete TestRun
        Write results of TestRun to output channel(s)
        """

        self.timing.takeTime(GC.TIMING_TESTRUN)
        self.timing.takeTimeSumOutput()

        if self.apiInstance:
            self.apiInstance.tearDown()

        network_info = self.browserFactory.teardown()
        self.kwargs['networkInfo'] = network_info

        if self.testCasesEndDateTimes_1D:
            self.kwargs['testCasesEndDateTimes_1D'] = self.testCasesEndDateTimes_1D

        if self.testCasesEndDateTimes_2D and self.testCasesEndDateTimes_2D[0]:
            self.kwargs['testCasesEndDateTimes_2D'] = self.testCasesEndDateTimes_2D

        if len(self.additionalExportTabs) > 0:
            self.kwargs[GC.EXPORT_ADDITIONAL_DATA] = self.additionalExportTabs

        self.results = ExportResults(**self.kwargs)  # -- API support: self.results --
        successful, error = self.getSuccessAndError()
        waiting = self.getWaiting()
        self.statistics.update_all(successful, error, waiting)
        logger.info(f"Finished execution of Testrun {self.testRunName}. "
                    f"{successful} Testcases successfully executed, {error} errors")
        print(f"Finished execution of Testrun {self.testRunName}. "
              f"{successful} Testcases successfully executed, {error} errors")

    def getSuccessAndError(self):
        """
        Returns number of successful and number of error test cases of the current test run
        @rtype: object
        """
        lError = 0
        lSuccess = 0
        for value in self.dataRecords.values():
            if value[GC.TESTCASESTATUS] == GC.TESTCASESTATUS_ERROR:
                lError += 1
            elif value[GC.TESTCASESTATUS] == GC.TESTCASESTATUS_SUCCESS:
                lSuccess += 1
        return lSuccess, lError

    def getWaiting(self):
        lWaiting =0
        for value in self.dataRecords.values():
            if value[GC.TESTCASESTATUS] == GC.TESTCASESTATUS_WAITING:
                lWaiting += 1
        return lWaiting

    def getAllTestRunAttributes(self):
        return self.testRunUtils.getCompleteTestRunAttributes(self.testRunName)

    def getBrowser(self, browserInstance=0, browserName=None, browserAttributes=None, mobileType=None, mobileApp=None,
                   desired_app=None, mobile_app_setting=None, browserWindowSize=None):

        return self.browserFactory.getBrowser(browserInstance=browserInstance,
                                              browserName=browserName,
                                              browserAttributes=browserAttributes,
                                              mobileType=mobileType,
                                              mobileApp=mobileApp,
                                              desired_app=desired_app,
                                              mobile_app_setting=mobile_app_setting,
                                              browserWindowSize=browserWindowSize)

    def getAPI(self):
        if not self.apiInstance:
            self.apiInstance = ApiHandling()
        return self.apiInstance

    def setResult(self, recordNumber, dataRecordResult):
        logger.debug(f"Received new result for Testrecord {recordNumber}")
        self.dataRecords[recordNumber] = dataRecordResult

    def executeTestSequence(self):
        """
        Start TestcaseSequence

        TestCaseSequence is a sequence of Testcases. In the TestcaseSequence there's a sequential List of
        Testcases to be executed.

        Before the loop (executeDictSequenceOfClasses) variables inside the the testrun-definition are replaced
        by values from the globals-file (e.g. if you want to generally run with FF, but in a certain case you want to
        run with Chrome, you'd have FF in the Testrundefinition, but set parameter in globals_chrome.json accordingly
        (in this case {"TC.Browser": "CHROME"}. TC.-Prefix signals the logic to look for this variable ("Browser")
        inside the testcase definitions and replace it with value "CHROME".

        """
        self.testRunUtils.replaceGlobals(self.globalSettings)
        self.testRunUtils.replaceClasses(self.testRunName, self.classesForObjects)

        kwargs = {GC.KWARGS_TESTRUNATTRIBUTES: self.getAllTestRunAttributes(),
                  GC.KWARGS_TESTRUNINSTANCE: self,
                  GC.KWARGS_TIMING: self.timing}

        self.executeDictSequenceOfClasses(
            kwargs[GC.KWARGS_TESTRUNATTRIBUTES][GC.STRUCTURE_TESTCASESEQUENCE],
            counterName=GC.STRUCTURE_TESTCASESEQUENCE, **kwargs)

    def executeDictSequenceOfClasses(self, dictSequenceOfClasses, counterName, **kwargs):
        """
        This is the main loop of the TestCaseSequence, TestCases, TestStepSequences and TestSteps.
        The Sequence of which class instance to create is defined by the TestRunAttributes.

        Before instancgetBrowsering the class it is checked, whether the class was loaded already and if not, will be loaded
        (only if the classname is fully qualified (e.g baangt<projectname>.TestSteps.myTestStep).
        If the testcase-Status is already "error" (GC.TESTCASESTATUS_ERROR) we'll stop the loop.

        @param dictSequenceOfClasses: The list of classes to be instanced. Must be a dict of {Enum, Classname},
        can be also {enum: [classname, <whatEverElse>]}
        @param counterName: Which Structure element we're currently looping, e.g. "TestStep" (GC.STRUCTURE_TESTSTEP)
        @param kwargs: TestrunAttributes, this TestRun, the Timings-Instance, the datarecord

        """
        if not kwargs.get(GC.KWARGS_TESTRUNATTRIBUTES):
            kwargs[GC.KWARGS_TESTRUNATTRIBUTES] = self.getAllTestRunAttributes()
        if not kwargs.get(GC.KWARGS_TESTRUNINSTANCE):
            kwargs[GC.KWARGS_TESTRUNINSTANCE] = self
            logger.info('get into not kwargs.getGC.KWARGS_TESTRUNINSTANCE, id is {}'.format(id(self)))
        if not kwargs.get(GC.KWARGS_TIMING):
            kwargs[GC.KWARGS_TIMING] = self.timing
        for key, value in dictSequenceOfClasses.items():
            # If any of the previous steps set the testcase to "Error" - exit here.
            if kwargs.get(GC.KWARGS_DATA):
                if kwargs[GC.KWARGS_DATA][GC.TESTCASESTATUS] == GC.TESTCASESTATUS_ERROR:
                    logger.info(f"TC is already in status Error - not processing steps {counterName}: {key}, {value}"
                                f"and everything behind this step")
                    return
                if kwargs[GC.KWARGS_DATA].get(GC.TESTCASESTATUS_STOP):
                    logger.info(f"TC wanted to stop. Not processing steps {counterName}: {key}, {value}"
                                f"and everything behind this step. TC-Status is "
                                f"{kwargs[GC.KWARGS_DATA][GC.TESTCASESTATUS]}")
                    return
                if kwargs[GC.KWARGS_DATA].get(GC.TESTCASESTATUS_STOPERROR):
                    kwargs[GC.KWARGS_DATA][GC.TESTCASESTATUS] = GC.TESTCASESTATUS_ERROR
                    if not kwargs[GC.KWARGS_DATA].get(GC.TESTCASEERRORLOG):
                        kwargs[GC.KWARGS_DATA][GC.TESTCASEERRORLOG] = "Aborted by command 'TCStopTestCaseError'"
                    else:
                        kwargs[GC.KWARGS_DATA][GC.TESTCASEERRORLOG] = kwargs[GC.KWARGS_DATA][GC.TESTCASEERRORLOG] \
                                                                      + "\nAborted by command 'TCStopTestCaseError'"
                    return

            logger.info(f"Starting {counterName}: {key}")
            kwargs[counterName] = key

            # Get the class reference:
            if isinstance(value, list):
                lFullQualified = value[0]  # First List-Entry must hold the ClassName
            else:
                lFullQualified = value

            l_class = TestRun.__dynamicImportClasses(lFullQualified)

            try:
                l_class(**kwargs)  # Executes the class´es  __init__ method
            except TypeError as e:
                # Damn! The class is wrong.
                l_class = TestRun.__dynamicImportClasses(lFullQualified)
                l_class(**kwargs)

        self.kwargs = kwargs

    def _initTestRunSettingsFromFile(self):
        self.__loadJSONGlobals()
        self.__setPathsIfNotPredefined()
        self.__sanitizeGlobalsValues()

    def __setPathsIfNotPredefined(self):
        if not self.globalSettings.get(GC.PATH_SCREENSHOTS, None):
            self.globalSettings[GC.PATH_SCREENSHOTS] = str(
                Path(self.managedPaths.getOrSetScreenshotsPath()).expanduser())
        else:
            self.managedPaths.getOrSetScreenshotsPath(path=self.globalSettings.get(GC.PATH_SCREENSHOTS))
        if not self.globalSettings.get(GC.PATH_EXPORT, None):
            self.globalSettings[GC.PATH_EXPORT] = str(Path(self.managedPaths.getOrSetExportPath()).expanduser())
        else:
            self.managedPaths.getOrSetExportPath(path=self.globalSettings.get(GC.PATH_EXPORT))
        if not self.globalSettings.get(GC.PATH_IMPORT, None):
            self.globalSettings[GC.PATH_IMPORT] = str(Path(self.managedPaths.getOrSetImportPath()).expanduser())
        else:
            self.managedPaths.getOrSetImportPath(path=self.globalSettings.get(GC.PATH_IMPORT))
        if not self.globalSettings.get(GC.PATH_ROOT, None):
            self.globalSettings[GC.PATH_ROOT] = str(Path(self.managedPaths.getOrSetRootPath()).parent.expanduser())
        else:
            self.managedPaths.getOrSetRootPath(path=self.globalSettings.get(GC.PATH_ROOT))

    def __loadJSONGlobals(self):
        if self.globalSettingsFileNameAndPath:
            self.globalSettings = utils.openJson(self.globalSettingsFileNameAndPath)

        # Set default execution STAGE
        if not self.globalSettings.get(GC.EXECUTION_STAGE, None):
            logger.debug(f"Execution Stage was not set. Setting to default value {GC.EXECUTION_STAGE_TEST}")
            self.globalSettings[GC.EXECUTION_STAGE] = GC.EXECUTION_STAGE_TEST

    def __sanitizeGlobalsValues(self):
        # Support for new dataClass to load different Classes
        for key, value in self.globalSettings.items():
            if "CL." in key:
                self.classesForObjects.__setattr__(key.strip("CL."), value)

            # Change boolean strings into boolean values.
            if isinstance(value, str):
                if value.lower() in ("false", "true", "no", "x"):
                    self.globalSettings[key] = utils.anything2Boolean(value)

            if isinstance(value, dict):
                if "default" in value:
                    # This happens in the new UI, if a value was added manually,
                    # but is not part of the globalSetting.json. In this case there's the whole shebang in a dict. We
                    # are only interested in the actual value, which is stored in "default":
                    self.globalSettings[key] = value["default"]
                    continue
                else:
                    # This could be the "old" way of the globals-file (with {"HEADLESS":"True"})
                    self.globalSettings[key] = value
                    continue

            if isinstance(value, str) and len(value) > 0:
                if value[0] == "{" and value[-1] == "}":
                    # Dict, that is not seen as dict
                    value = value.replace("\'", '"')
                    self.globalSettings[key] = json.loads(value)

        if self.globalSettings.get("TC." + GC.EXECUTION_LOGLEVEL):
            utils.setLogLevel(self.globalSettings.get("TC." + GC.EXECUTION_LOGLEVEL))

    def _loadJSONTestRunDefinitions(self):
        if not self.testRunFileName and not self.testRunDict:  # -- API support: testRunDict --
            return

        if self.testRunFileName and ".JSON" in self.testRunFileName.upper():  # -- API support: self.testRunFileName --
            data = utils.replaceAllGlobalConstantsInDict(utils.openJson(self.testRunFileName))
            self.testRunUtils.setCompleteTestRunAttributes(testRunName=self.testRunName,
                                                           testRunAttributes=data)

        # -- API support --
        # load TestRun from dict
        if self.testRunDict:
            data = utils.replaceAllGlobalConstantsInDict(self.testRunDict)
            self.testRunUtils.setCompleteTestRunAttributes(testRunName=self.testRunName,
                                                           testRunAttributes=data)
        # -- END of API support --

    def _loadExcelTestRunDefinitions(self):
        if not self.testRunFileName:
            return

        if ".XLSX" in self.testRunFileName.upper():
            logger.info(f"Reading Definition from {self.testRunFileName}")
            lExcelImport = TestRunExcelImporter(FileNameAndPath=self.testRunFileName, testRunUtils=self.testRunUtils)
            lExcelImport.importConfig(global_settings=self.globalSettings)

    @staticmethod
    def __dynamicImportClasses(fullQualifiedImportName):
        return utils.dynamicImportOfClasses(fullQualifiedImportName=fullQualifiedImportName)

    @staticmethod
    def _sanitizeTestRunNameAndFileName(TestRunNameInput):
        """
        @param TestRunNameInput: The complete File and Path of the TestRun definition (JSON or XLSX).
        @return: TestRunName and FileName (if definition of testrun comes from a file (JSON or XLSX)
        """
        if ".XLSX" in TestRunNameInput.upper() or ".JSON" in TestRunNameInput.upper():
            lRunName = utils.extractFileNameFromFullPath(TestRunNameInput)
            lFileName = TestRunNameInput
        else:
            lRunName = TestRunNameInput
            lFileName = None

        return lRunName, lFileName




import os
import csv
import xlrd3 as xlrd
import json
import xlsxwriter
import logging
import requests
import pymsteams
import configparser
from slack_webhook import Slack
from baangt.base.PathManagement import ManagedPaths
from baangt.base.SendReports.mailer import SendMail
from baangt.base.ExportResults.ExportConfluence import ExportConfluence

logger = logging.getLogger('pyC')


class Sender:
    def __init__(self, settings, filename, attach_file=""):
        self.defaultSettings = settings
        self.xlsx_file = filename
        self.attach_file = attach_file
        self.subject, self.body = self.create_body()

    def sendMail(self):
        self.set_globalSettings(self.defaultSettings, "SendMailTo")
        if len(self.globalSettings.get("SendMailTo")) > 0:
            SendMail(self.globalSettings, self.xlsx_file, self.subject, self.body, self.attach_file)

    def sendMsTeam(self):
        self.set_globalSettings(self.defaultSettings, "MsWebHook")
        if len(self.globalSettings.get("MsWebHook")) > 0:
            lis = self.get_list(self.globalSettings.get("MsWebHook"))
            try:
                for hooks in lis:
                    text = "<pre>" + self.subject + "\n\n" + self.body + "</pre>"
                    text.replace('\n', '<br>')
                    msg = pymsteams.connectorcard(hooks)
                    msg.text(text)
                    msg.send()
                    logger.info(f"MsTeam sent to: {hooks}")
            except Exception as ex:
                logger.info(ex)

    def sendSlack(self):
        self.set_globalSettings(self.defaultSettings, "SlackWebHook")
        if len(self.globalSettings.get("SlackWebHook")) > 0:
            lis = self.get_list(self.globalSettings.get("SlackWebHook"))
            try:
                for hooks in lis:
                    text = self.subject+"\n\n"+self.body
                    slack = Slack(url=hooks)
                    slack.post(text=text)
                    logger.info(f"Slack sent to: {hooks}")
            except Exception as ex:
                logger.info(ex)

    def sendTelegram(self, test=False):
        self.set_globalSettings(self.defaultSettings, "TelegramBot")
        bot = self.globalSettings.get("TelegramBot")
        if len(bot) > 0 and len(self.globalSettings.get("TelegramChannel")):
            lis = self.get_list(self.globalSettings.get("TelegramChannel"))
            messages = []
            for channel in lis:
                try:
                    text = self.subject + "\n\n" + self.body
                    if channel[0] != '@':
                        channel = '@' + channel
                    url = f"https://api.telegram.org/bot{bot}/sendMessage?chat_id={channel}&text={text}"
                    res = requests.get(url)
                    js = json.loads(res.content)
                    if res.status_code == 200:
                        logger.info(f"Sent to {channel} telegram channel")
                        if js['ok']:
                            messages.append(js["result"]["text"])
                    else:
                        logger.info(res.content)
                except Exception as ex:
                    logger.info(ex)
            if test:
                return messages

    def sendConfluence(self):
        self.set_globalSettings(self.defaultSettings, "Confluence-Base-Url")
        if self.globalSettings["Confluence-Base-Url"]:
            args = {}
            args['url'] = self.globalSettings["Confluence-Base-Url"]
            args['space'] = self.globalSettings["Confluence-Space"]
            args['pageTitle'] = self.globalSettings["Confluence-Pagetitle"]
            args['fileNameAndPathToResultXLSX'] = self.xlsx_file
            args['username'] = self.globalSettings["Confluence-Username"]
            args['password'] = self.globalSettings["Confluence-Password"]
            args['rootPage'] = self.globalSettings["Confluence-Rootpage"]
            args['remove_headers'] = [x.strip() for x in self.globalSettings["Confluence-Remove_Headers"].split(",")]
            args['uploadOriginalFile'] = self.globalSettings["Confluence-Uploadoriginalfile"]
            args['CreateSubPagesForEachXXEntries'] = int(self.globalSettings["Confluence-Createsubpagesforeachxxentries"])
            ExportConfluence(**args)

    def set_globalSettings(self, setting, word_to_look):
        if word_to_look in setting:
            self.globalSettings = setting
        else:
            self.readConfig()

    def readConfig(self):
        """ Read existing mail.ini file if not exist then will write one with empty values"""
        self.globalSettings = {}
        config = configparser.ConfigParser()
        managedPaths = ManagedPaths()
        config_file = managedPaths.getOrSetIni().joinpath("mail.ini")
        if not os.path.exists(config_file):
            self.write_config(config_file)
        config.read(config_file)
        self.globalSettings["SendMailTo"] = config["Default"].get("SendMailTo"
                                            ) or self.write_config(config_file, "SendMailTo", "")
        self.globalSettings["NotificationWithAttachment"] = config["Default"].get("NotificationWithAttachment"
                                            ) or self.write_config(config_file, "NotificationWithAttachment", False)
        self.globalSettings["MsWebHook"] = config["Default"].get("MsWebHook"
                                            ) or self.write_config(config_file, "MsWebHook", "")
        self.globalSettings["SlackWebHook"] = config["Default"].get("SlackWebHook"
                                            ) or self.write_config(config_file, "SlackWebHook", "")
        self.globalSettings["TelegramBot"] = config["Default"].get("TelegramBot"
                                                                    ) or self.write_config(config_file, "TelegramBot",
                                                                                           "")
        self.globalSettings["TelegramChannel"] = config["Default"].get("TelegramChannel"
                                                                    ) or self.write_config(config_file, "TelegramChannel",
                                                                                           "")
        self.globalSettings["Confluence-Base-Url"] = config["Default"].get("Confluence-Base-Url") or self.write_config(
            config_file, "Confluence-Base-Url", "")

    def write_config(self, config_file, key=None, value=None):
        if key:
            config_write = configparser.ConfigParser()
            config_write["Default"] = {key: value}
            for keys in self.globalSettings:
                config_write["Default"][keys] = self.globalSettings[keys]
            with open(config_file, "w") as configFile:
                config_write.write(configFile)
            return value
        config_write = configparser.ConfigParser()
        config_write["Default"] = {
            "SendMailTo": "",
            "NotificationWithAttachment": 'False',
            "MsWebHook": "",
            "SlackWebHook": "",
            "TelegramBot": "",
            "TelegramChannel": ""
        }
        with open(config_file, "w") as configFile:
            config_write.write(configFile)

    def create_body(self):
        """
        Will create body from mail by using summary tab of xlsx output file.
        Any changes in body structure should be made in this function.
        :return:
        """
        wb = xlrd.open_workbook(self.xlsx_file)
        sheet = wb.sheet_by_name("Summary")
        data = []
        for x in range(2, sheet.nrows):
            lis = []
            for cell in sheet.row(x):
                value = cell.value
                if type(value) == float:
                    value = str(value)
                    if value[-2:] == '.0':
                        value = value[:-2]
                else:
                    value = str(value).strip()
                lis.append(value)
            while "" in lis:
                lis.remove("")
            if lis != []:
                data.append(' = '.join(lis))
        subject = sheet.row(0)[0].value.strip()
        body = '\n'.join(data)
        return subject, body

    def get_list(self, string):
        lis = string.split(',')
        lis = [data.strip() for data in lis]
        return lis

    @staticmethod
    def send_all(results, globalSettings):
        if ".csv" in results.fileName:
            # If output file is of CSV Format then we are creating a temporary xlsx file which is just used to
            # send reports and get deleted after that.
            temp_file = results.fileName + ".xlsx"
            results.workbook = xlsxwriter.Workbook(temp_file)
            results.summarySheet = results.workbook.add_worksheet("Summary")
            results.cellFormatGreen = results.workbook.add_format()
            results.cellFormatGreen.set_bg_color('green')
            results.cellFormatRed = results.workbook.add_format()
            results.cellFormatRed.set_bg_color('red')
            results.cellFormatBold = results.workbook.add_format()
            results.cellFormatBold.set_bold(bold=True)
            results.summaryRow = 0
            results.makeSummaryExcel()
            results.closeExcel()
            send_stats = Sender(globalSettings, temp_file, results.fileName)
            os.remove(temp_file)
        else:
            send_stats = Sender(globalSettings, results.fileName)

        try:
            send_stats.sendMail()
        except Exception as ex:
            logger.error(f"Error when sending mail: {ex}")

        try:
            send_stats.sendMsTeam()
        except Exception as ex:
            logger.error(f"Error when sending MS Teams: {ex}")

        try:
            send_stats.sendSlack()
        except Exception as ex:
            logger.error(f"Error when sending Slack: {ex}")

        try:
            send_stats.sendTelegram()
        except Exception as ex:
            logger.error(f"Error when sending Telegram: {ex}")

        try:
            send_stats.sendConfluence()
        except Exception as ex:
            logger.error(f"Error when saving in Confluence: {ex}")





from datetime import datetime
from sqlalchemy.orm import sessionmaker
from sqlalchemy import desc, and_
from baangt.base.DataBaseORM import engine, TestrunLog, GlobalAttribute, TestCaseLog, TestCaseSequenceLog, TestCaseField
import baangt.base.GlobalConstants as GC
from jinja2 import Environment, FileSystemLoader
import json
import os
import webbrowser
import uuid

# number of items on history charts
history_items = 10
template_dir = 'templates'

class Report:
	#
	# Parent reports class
	# defines constants and construcor
	#

	def __init__(self):
		self.created = datetime.now()
		self.generate();

	@property
	def path(self):
		return ''	

	def generate(self):
		#
		# report generator
		#

		pass

	def show(self):
		#
		# shows html report
		#

		webbrowser.open(f'file://{self.path}', new=2)


	def template(self, template_name):
		#
		# returns jinja2 template
		#

		file_loader = FileSystemLoader(os.path.join(os.path.dirname(__file__), GC.REPORT_PATH, template_dir))
		env = Environment(loader=file_loader)

		return env.get_template(template_name)


	def chart_figures(self, log):
		#
		# builds json with log statistics
		#

		return {
			'records': log.recordCount,
			'successful': log.statusOk,
			'error': log.statusFailed,
			'paused': log.statusPaused,
		}

	def time_in_seconds(self, time):
		time_int = time.split('.')[0]
		factors = [3600, 60, 1]
		return sum(t*f for t,f in zip(map(int, time_int.split(':')), factors))

	def chart_status(self, log):
		#
		# builds json for Chart.js: log statistics
		#

		return json.dumps({
			'type': 'doughnut',
			'data': {
				'datasets': [{
					'data': [
						log.statusOk,
						log.statusFailed,
						log.statusPaused,
					],
					'backgroundColor': [
						'#52ff52',
						'#ff5252',
						'#bdbdbd',
					],
				}],
				'labels': [
					'Passed',
					'Failed',
					'Paused'
				],
			},
			'options': {
				'legend': {
					'display': False,
				},
			},
		})

	def chart_results(self, logs):
		#
		# builds json for Chart.js: history of log results
		#

		logs = logs[-history_items:]
		empty = history_items - len(logs)

		return json.dumps({
			'type': 'bar',
				'data': {
					'datasets': [
						{
							'data': [x.statusOk for x in logs] + [None]*empty,
							'backgroundColor': '#52ff52',
							'label': 'Passed',
						},
						{
							'data': [x.statusFailed for x in logs] + [None]*empty,
							'backgroundColor': '#ff5252',
							'label': 'Failed',
						},
						{
							'data': [x.statusPaused for x in logs] + [None]*empty,
							'backgroundColor': '#bdbdbd',
							'label': 'Paused',
						},
					],
					'labels': [x.startTime.strftime('%Y-%m-%d %H:%M') for x in logs] + [None]*empty,
				},
				'options': {
					'legend': {
						'display': False,
					},
					'tooltips': {
						'mode': 'index',
						'intersect': False,
					},
					'scales': {
						'xAxes': [{
							'gridLines': {
								'display': False,
								'drawBorder': False,
							},
							'ticks': {
								'display': False,
							},
							'stacked': True,
						}],
						'yAxes': [{
							'gridLines': {
								'display': False,
								'drawBorder': False,
							},
							'ticks': {
								'display': False,
							},
							'stacked': True,
						}],
					},
				},
			})

	def chart_duration(self, logs):
		#
		# builds json for Chart.js: history of log durations
		#

		logs = logs[-history_items:]
		empty = history_items - len(logs)

		return json.dumps({
			'type': 'line',
				'data': {
					'datasets': [{
						'data': [str(x.duration) for x in logs] + [None]*empty,
						'fill': False,
						'borderColor': '#cfd8dc',
						'borderWidth': 1, 
						'pointBackgroundColor': 'transparent',
						'pointBorderColor': '#0277bd',
						'pointRadius': 5,
						'lineTension': 0,
						'label': 'Duration',
					}],
					'labels': [x.startTime.strftime('%Y-%m-%d %H:%M') for x in logs] + [None]*empty,
				},
				'options': {
					'legend': {
						'display': False,
					},
					'tooltips': {
						'mode': 'index',
						'intersect': False,
					},
					'layout': {
						'padding': {
							'top': 10,
							'right': 10,
						},
					},
					'scales': {
						'xAxes': [{
							'gridLines': {
								'display': False,
								'drawBorder': False,
							},
							'ticks': {
								'display': False,
							},
						}],
						'yAxes': [{
							'gridLines': {
								'display': False,
								'drawBorder': False,
							},
							'ticks': {
								'display': False,
							},
						}],
					},
				},
			})


	def chart_testcases(self, logs):
		#
		# builds json for Chart.js: history of log results
		#

		tc = logs[-history_items:]
		empty = history_items - len(tc)

		return json.dumps({
			'type': 'bar',
				'data': {
					'datasets': [
						{
							'data': [self.time_in_seconds(t['duration']) if t['status'] == GC.TESTCASESTATUS_SUCCESS else 0 for t in tc] + [None]*empty,
							'backgroundColor': '#52ff52',
							'label': 'PASSED',
						},
						{
							'data': [self.time_in_seconds(t['duration']) if t['status'] == GC.TESTCASESTATUS_ERROR else 0 for t in tc] + [None]*empty,
							'backgroundColor': '#ff5252',
							'label': 'FAILED',
						},
						{
							'data': [self.time_in_seconds(t['duration']) if t['status'] == GC.TESTCASESTATUS_WAITING else 0 for t in tc] + [None]*empty,
							'backgroundColor': '#bdbdbd',
							'label': 'PAUSED',
						},
					],
					'labels': [t['duration'] for t in tc] + [None]*empty,
				},
				'options': {
					'legend': {
						'display': False,
					},
					'tooltips': {
						'mode': 'index',
						'intersect': False,
					},
					'scales': {
						'xAxes': [{
							'gridLines': {
								'display': False,
								'drawBorder': False,
							},
							'ticks': {
								'display': False,
							},
							'stacked': True,
						}],
						'yAxes': [{
							'gridLines': {
								'display': False,
								'drawBorder': False,
							},
							'ticks': {
								'display': False,
							},
							'stacked': True,
						}],
					},
				},
			})



class Dashboard(Report):

	def __init__(self, name=None, stage=None):
		self.name = name
		self.stage = stage
		super().__init__()


	@property
	def path(self):
		#
		# path to report
		#
		return os.path.join(
			os.path.dirname(__file__),
			GC.REPORT_PATH,
			self.created.strftime('dashboard-%Y%m%d-%H%M%S.html'),
		)


	def generate(self):
		#
		# generates HTML report
		#

		# get jinja2 template
		template = self.template('dashboard.html')

		data = self.get_data() or None
		if data is None:
			error_msg = 'No TestrunLog maches:'
			if self.name:
				error_msg = f'{error_msg} Name "{self.name}"'
				if self.stage:
					  error_msg = f'{error_msg},'
			if self.stage:
				error_msg = f'{error_msg} Stage "{self.stage}"'
			raise ValueError(error_msg)

		# generate report
		with open(self.path, 'w') as f:
			f.write(template.render(
				type='Dashboard',
				data=data,
				created=self.created.strftime('%Y-%m-%d %H:%M:%S'),
				name=self.name,
				stage=self.stage,
			))


	def get_data(self):
		#
		# get data from db for the report
		#

		db = sessionmaker(bind=engine)()
		records = []

		if self.name and self.stage:
			logs = db.query(TestrunLog).order_by(TestrunLog.startTime).filter_by(testrunName=self.name)\
				.filter(TestrunLog.globalVars.any(and_(GlobalAttribute.name==GC.EXECUTION_STAGE, GlobalAttribute.value==self.stage))).all()
			return {'records': [self.build_charts(logs)]}
		
		elif self.name:
			# get Testrun stages
			stages = db.query(GlobalAttribute.value).filter(GlobalAttribute.testrun.has(TestrunLog.testrunName==self.name))\
			.filter_by(name=GC.EXECUTION_STAGE).group_by(GlobalAttribute.value).order_by(GlobalAttribute.value).all()
			stages = [x[0] for x in stages]

			for stage in stages:
				logs = db.query(TestrunLog).order_by(TestrunLog.startTime).filter_by(testrunName=self.name)\
					.filter(TestrunLog.globalVars.any(and_(GlobalAttribute.name==GC.EXECUTION_STAGE, GlobalAttribute.value==stage))).all()
				records.append(self.build_charts(logs, stage=stage))

			return {'records': records}

		elif self.stage:
			# get Testrun names
			names = db.query(TestrunLog.testrunName)\
			.filter(TestrunLog.globalVars.any(and_(GlobalAttribute.name==GC.EXECUTION_STAGE, GlobalAttribute.value==self.stage)))\
			.group_by(TestrunLog.testrunName).order_by(TestrunLog.testrunName).all()
			names = [x[0] for x in names]

			for name in names:
				logs = db.query(TestrunLog).order_by(TestrunLog.startTime).filter_by(testrunName=name)\
					.filter(TestrunLog.globalVars.any(and_(GlobalAttribute.name==GC.EXECUTION_STAGE, GlobalAttribute.value==self.stage))).all()
				records.append(self.build_charts(logs, name=name))

			return {'records': records}

		else:
			# get Testrun names
			names = db.query(TestrunLog.testrunName)\
			.group_by(TestrunLog.testrunName).order_by(TestrunLog.testrunName).all()
			names = [x[0] for x in names]
			stage_set = set()

			for name in names:
				# get Testrun stages
				stages = db.query(GlobalAttribute.value).filter(GlobalAttribute.testrun.has(TestrunLog.testrunName==name))\
				.filter_by(name=GC.EXECUTION_STAGE).group_by(GlobalAttribute.value).order_by(GlobalAttribute.value).all()
				stages = [x[0] for x in stages]
				stage_set.update(stages)

				for stage in stages:
					logs = db.query(TestrunLog).order_by(TestrunLog.startTime).filter_by(testrunName=name)\
						.filter(TestrunLog.globalVars.any(and_(GlobalAttribute.name==GC.EXECUTION_STAGE, GlobalAttribute.value==stage))).all()
					records.append(self.build_charts(logs, name=name, stage=stage))

			return {
				'names': names,
				'stages': list(stage_set),
				'records': records,
			}


	def build_charts(self, logs, name=None, stage=None):
		#
		# builds Chart.js data collections
		#

		name = self.name or name
		stage = self.stage or stage
		
		if logs:
			return {
				'id': f'record-{uuid.uuid4()}',
				'name': name,
				'stage': stage,
				'figures': self.chart_figures(logs[-1]),
				'status': self.chart_status(logs[-1]),
				'results': self.chart_results(logs),
				'duration': self.chart_duration(logs),
			}
		else:
			error_msg = 'No TestrunLog maches:'
			if self.name:
				error_msg = f'{error_msg} Name "{self.name}"'
				if self.stage:
					  error_msg = f'{error_msg},'
			if self.stage:
				error_msg = f'{error_msg} Stage "{self.stage}"'
			raise ValueError(error_msg)



class Summary(Report):

	def __init__(self, id):
		self.id = id
		super().__init__()

	@property
	def path(self):
		#
		# path to report
		#
		return os.path.join(
			os.path.dirname(__file__),
			GC.REPORT_PATH,
			self.created.strftime('summary-%Y%m%d-%H%M%S.html'),
		)

	def generate(self):
		#
		# generates HTML report
		#

		# get jinja2 template
		template = self.template('summary.html')

		# generate report
		with open(self.path, 'w') as f:
			f.write(template.render(
				type='Report',
				data=self.get_data(),
				created=self.created.strftime('%Y-%m-%d %H:%M:%S'),
			))

	def get_data(self):
		#
		# get data from db for the report
		#

		db = sessionmaker(bind=engine)()

		log = db.query(TestrunLog).get(uuid.UUID(self.id).bytes)
		if log is None:
			raise ValueError(f'TestrunLog {self.id} does not exist')

		# collect TestCases and screenshots
		testcases = []
		screenshots = []
		for index, tc in enumerate(db.query(TestCaseLog).filter(TestCaseLog.testcase_sequence.has(TestCaseSequenceLog.testrun_id == log.id))):
			testcases.append({
				'status': db.query(TestCaseField.value).filter_by(name=GC.TESTCASESTATUS).filter(TestCaseField.testcase_id == tc.id).first()[0],
				'duration': db.query(TestCaseField.value).filter_by(name=GC.TIMING_DURATION).filter(TestCaseField.testcase_id == tc.id).first()[0],
			})
			tc_screenshots = db.query(TestCaseField.value).filter_by(name=GC.SCREENSHOTS).filter(TestCaseField.testcase_id == tc.id).first()[0]
			if tc_screenshots:
				for shot in json.loads(tc_screenshots.replace("'", '"')):
					screenshots.append({
						'index': index + 1,
						'path': shot,
					})

		# collect files
		files = [
			{
				'name': 'Log',
				'path': log.logfileName,
			},
			{
				'name': 'Results',
				'path': log.dataFile,
			}
		]

		data = {
			'id': self.id,
			'time': log.startTime.strftime('%Y-%m-%d %H:%M:%S'),
			'name': log.testrunName,
			'figures': self.chart_figures(log),
			'status': self.chart_status(log),
			'testcases': self.chart_testcases(testcases),
			'screenshots': screenshots,
			'files': files,
		}

		return data




		


import json
import logging
import requests
from dataclasses import dataclass, asdict, field
from typing import Dict
from baangt.base.RuntimeStatistics import Statistic


logger = logging.getLogger("pyC")


class Singleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


def default_field(obj):
    return field(default_factory=lambda: obj)

@dataclass
class Statistics(metaclass=Singleton):
    Activity: Dict[str, int] = default_field({})
    # TestCaseType
    Browser: int = 0
    api: int = 0
    soap: int = 0
    web: int = 0

    SequenceClass: str = ""
    TestCaseExecuted: int = 0
    TestCasePassed: int = 0
    TestCaseFailed: int = 0
    TestCasePaused: int = 0
    TestCaseClass: str = ""
    TestStepMaster: Dict[str, int] = default_field({})
    TestStepSequences: int = 0
    TestSteps: int = 0
    TestCaseSequences: int = 0
    # LocatorType
    XPATH: int = 0
    CSS: int = 0
    ID: int = 0
    # Comparision
    equalTo: int = 0
    notEqualTo: int = 0
    greaterThan: int = 0
    smallerThan: int = 0
    # Non-empty Values count
    timeout: int = 0
    optional: int = 0
    release: int = 0
    Duration: str = ""
    TestRunUUID: str = ""
    comparision_dict = {"=": "equalTo", "!=": "notEqualTo", ">": "greaterThan", "<": "smallerThan"}

    def update_data(self, dic):
        self.SequenceClass = Statistics.get_value(dic, "SequenceClass")
        self.TestCaseClass = Statistics.get_value(dic, "TestCaseClass")
        TestCaseType = Statistics.get_value(dic, "TestCaseType")
        TestStep = Statistics.get_value(dic, "TestStepExecutionParameters")
        TestStepMaster_lis = Statistics.get_value(dic, "TestStepClass", lis=True)
        for value in TestStepMaster_lis:
            if value not in self.TestStepMaster:
                self.TestStepMaster[value] = 1
            else:
                self.TestStepMaster[value] += 1

        if len(TestCaseType) > 0:
            self.update_attribute(TestCaseType)

        if len(self.SequenceClass) == 0:
            try:
                self.SequenceClass = Statistics.get_value(dic, "TESTSEQUENCE")[1][0]
            except KeyError:
                try:
                    self.SequenceClass = Statistics.get_value(dic, "TESTSEQUENCE")['1'][0]
                except KeyError:
                    pass

        if len(self.TestCaseClass) == 0:
            try:
                self.TestCaseClass = Statistics.get_value(dic, "TESTCASE")[1][0]
            except KeyError:
                try:
                    self.TestCaseClass = Statistics.get_value(dic, "TESTCASE")['1'][0]
                except KeyError:
                    pass

        for key in TestStep:
            self.update_attribute(TestStep[key]["Activity"].upper(), prefix="Activity_")
            if len(TestStep[key]["LocatorType"]) > 0:
                self.update_attribute(TestStep[key]["LocatorType"].upper())
            if len(TestStep[key]["Comparison"]) > 0:
                self.update_attribute(self.comparision_dict[TestStep[key]["Comparison"]])
            if len(str(TestStep[key]["Timeout"])) > 0:
                self.update_attribute("timeout")
            if len(TestStep[key]["Optional"]) > 0:
                self.update_attribute("optional")
            if len(TestStep[key]["Release"]) > 0:
                self.update_attribute("release")


    def update_attribute(self, string, prefix=""):
        string = prefix+string
        if prefix == "Activity_":
            if string not in self.Activity:
                self.Activity[string] = 0
            self.Activity[string] += 1
        else:
            var = getattr(self, string)
            var += 1
            setattr(self, string, var)

    def update_attribute_with_value(self, string, value):
        setattr(self, string, value)
        
    def update_runtimeStatistic(self):
        runtimeStats = Statistic()
        self.TestCaseSequences = runtimeStats.testcase_sequence_executed
        self.TestSteps = runtimeStats.teststep_executed
        self.TestStepSequences = runtimeStats.teststep_sequence_executed

    def to_dict(self):
        removeable = []
        dic = asdict(self)
        new_dic = {}
        for key in dic:
            if key[0] == "-":
                removeable.append(key)
            if dic[key] == 0 or dic[key] == "":
                removeable.append(key)
            if key == "Activity":
                for ky in dic[key]:
                    new_dic[ky] = dic[key][ky]
                removeable.append(key)
        for key in dic:
            if key not in removeable:
                new_dic[key] = dic[key]
        return new_dic

    def send_statistics(self, test=False):
        payload = self.to_dict()
        res = requests.post("https://stats.baangt.org", json=payload)
        if test:
            return res
        logger.debug(f"Statistics sent to server = {json.dumps(payload)}")

    @staticmethod
    def recursive_items(dictionary):
        for key, value in dictionary.items():
            if type(value) is dict:
                yield key, value
                yield from Statistics.recursive_items(value)
            if type(value) is list:
                yield key, value
                for val in value:
                    if type(val) is dict:
                        yield from Statistics.recursive_items(val)
            else:
                yield key, value

    @staticmethod
    def get_value(dictionary, get_key, lis=False):
        if lis:
            dic_lis = []
        for key, value in Statistics.recursive_items(dictionary):
            if key == get_key:
                if lis:
                    dic_lis.append(value)
                else:
                    return value
        if lis:
            return dic_lis
        return ""



from dataclasses import dataclass

@dataclass
class ClassesForObjects:
    browserFactory: str = "baangt.base.BrowserFactory.BrowserFactory"
    browserHandling: str = "baangt.base.BrowserHandling.BrowserHandling.BrowserDriver"
    testCaseSequenceMaster: str = "baangt.TestCaseSequence.TestCaseSequenceMaster"
    testStepMaster: str = "baangt.TestSteps.TestStepMaster"




import platform
import os
import subprocess
import xl2dict
import baangt.base.GlobalConstants as GC
from logging import getLogger

logger = getLogger("pyC")

def open(filenameAndPath: str):
    """Will open the given file with the Operating system default program.

    param

    filenameAndPath : Complete absolute path to file.
    return : True if sucessfully open. False if file doesn't exists or operating system doesn't have default program.
    """

    if not isinstance(filenameAndPath, str):
        filenameAndPath = str(filenameAndPath)

    filenameAndPath = os.path.abspath(filenameAndPath)
    logger.debug(f"Trying to open file with it's application: {filenameAndPath}")

    if not os.path.exists(filenameAndPath):
        logger.warning(f"Filename doesn't exist and can't be opened: {filenameAndPath}")
        return False

    elif platform.system().lower() == GC.PLATFORM_WINDOWS:
        try:
            filenameAndPath = f'"{filenameAndPath}"'
            os.startfile(filenameAndPath)
            return True
        except Exception as errorcode:
            if errorcode.errno == 22:
                os.popen(r"Rundll32.exe SHELL32.DLL, OpenAs_RunDLL "+filenameAndPath)
                return True
            else:
                return False

    elif platform.system().lower() == GC.PLATFORM_LINUX:
        logger.debug(f"In Linux trying to call xdg-open with filename: {str(filenameAndPath)}")
        status = subprocess.call(["xdg-open", str(filenameAndPath)])
        if status == 0:
            return True
        else:
            return False

    elif platform.system().lower() == GC.PLATFORM_MAC:
        filenameAndPath = f'"{filenameAndPath}"'
        status = os.system("open " + str(filenameAndPath))
        if status == 0:
            return True
        else:
            return False


class FilesOpen:
    """
    Class is called from UI and will open the corresponding file-type in the os-specific application.
    """
    def __init__(self):
        pass

    @staticmethod
    def openTestRunDefinition(filenameAndPath):
        try:
            xl_object = xl2dict.XlToDict()
            xl_dict = xl_object.fetch_data_by_column_by_sheet_name(filenameAndPath, sheet_name="TestCaseSequence")
            for data in xl_dict:
                file_path = data["TestDataFileName"]
                if not os.path.exists(file_path):
                    file_path = os.path.join(os.path.dirname(filenameAndPath), file_path)
                status = open(file_path)
        except:
            pass
        return open(filenameAndPath=filenameAndPath)

    @staticmethod
    def openResultFile(filenameAndPath):
        return open(filenameAndPath=filenameAndPath)

    @staticmethod
    def openLogFile(filenameAndPath):
        return open(filenameAndPath=filenameAndPath)









from atlassian import Confluence
from html import escape
import xlrd3 as xlrd
import os


class ExportConfluence:
    def __init__(self, url, space, pageTitle, fileNameAndPathToResultXLSX, username, password, rootPage=None,
                 remove_headers=["json"], uploadOriginalFile=False, CreateSubPagesForEachXXEntries=0):
        self.url = url
        self.space = space
        self.rootPage = rootPage
        self.pageTitle = pageTitle
        self.fileNameAndPathToResultXLSX = fileNameAndPathToResultXLSX
        self.username = username
        self.password = password
        self.remove_headers = [headers.lower() for headers in remove_headers]
        self.uploadOriginalFile = uploadOriginalFile
        self.CreateSubPagesForEachXXEntries = CreateSubPagesForEachXXEntries
        self.html = self.makeBody()
        self.update_confluence()

    def makeBody(self):
        summary = self.xlsx2html(self.fileNameAndPathToResultXLSX, sheet="Summary")
        if not self.CreateSubPagesForEachXXEntries:  # If not subPages then create a single page
            output = self.xlsx2html(self.fileNameAndPathToResultXLSX, sheet="Output")
            html = "<h1>Summary</h1>" + summary + "<br /><br /><h1>Output</h1>" + output  # joining output in main page
        else:
            html = "<h1>Summary</h1>" + summary + "<br />"  # Main page without output tab data
        html = html.replace('\\',  '\\\\')
        return html

    def update_confluence(self):
        confluence = Confluence(url=self.url, username=self.username, password=self.password)  # Confluence login
        if self.uploadOriginalFile:  # if original xlsx_file is to be attach on the page
            file = self.attach_file(confluence)
            html = file + "<br /><br />" + self.html
        else:
            html = self.html
        new_page = confluence.create_page(
            self.space, self.pageTitle, html, parent_id=self.rootPage, type='page', representation='storage'
        )  # creating main page
        if self.CreateSubPagesForEachXXEntries:  # if we want subpages for output
            try:  # getting page if of main page. With help of this id, sub pages are created
                parent_page = new_page["id"]
            except KeyError: # if key is not present then it is most probably inside a list with key "results"
                parent_page = new_page["results"]["id"]
            self.create_child_pages(confluence, parent_page)  # creating child pages

    def xlsx2html(self, filePath, sheet):
        wb = xlrd.open_workbook(filePath)
        sht = wb.sheet_by_name(sheet)
        data = []  # used to store rows, which are later joined to make complete html
        remove_index = []  # used to store columns which are removable
        header = []
        for row in range(sht.nrows):
            dt = []
            for column in range(sht.ncols):
                value = sht.cell_value(row, column)
                if type(value) == float:
                    if repr(value)[-2:] == '.0':
                        value = int(value)
                value = str(value)
                if row == 0:
                    if value.lower() in self.remove_headers:
                        remove_index.append(column)  # storing column number of removable headers in a list
                    else:
                        header.append(escape(value))
                else:
                    if column not in remove_index:  # if column is not in remove_header list than add the data in html
                        dt.append(escape(value))
            data.append('<td>' + '</td>\n<td>'.join(dt) + '</td>')  # joining individual data of a row in single row
        header = '<tr><th>' + '</th>\n<th>'.join(header) + '</th></tr>'
        html = '<table><tbody>' + header + '<tr>' + '</tr>\n<tr>'.join(data) + '</tr>' + '</tbody></table>'  # joining list of rows to make html
        return html

    def attach_file(self, confluence):
        fileName = os.path.basename(self.fileNameAndPathToResultXLSX)  # getting basename of xlsx file for title
        # Attaching basefile
        attach = confluence.attach_file(self.fileNameAndPathToResultXLSX, name=fileName, content_type=None,
                               page_id=self.rootPage, title=self.pageTitle, space=self.space, comment=None)
        try:
            url = "/confluence"+attach["_links"]["download"].split(".xlsx?")[0] + ".xlsx"
        except KeyError:  # if key is not present then it is most probably inside a list with key "results"
            url = "/confluence"+attach["results"][0]["_links"]["download"].split(".xlsx?")[0] + ".xlsx"
        link = f'<a href="{url}">{fileName}</a>'
        html = "<h1>Original file</h1>"+link
        return html

    def create_child_pages(self, confluence, parent_page):
        # Creating child pages if required
        wb = xlrd.open_workbook(self.fileNameAndPathToResultXLSX)
        output_xlsx = wb.sheet_by_name("Output")
        # Getting starting points for each subpage
        starting_points = [x for x in range(1, output_xlsx.nrows, self.CreateSubPagesForEachXXEntries)]
        header = []  # header used in every subpage
        remove_index = []  # headers column number which are not to be used are stored in it.
        for x in range(output_xlsx.ncols):
            value = output_xlsx.cell_value(0, x)
            if value.lower() not in self.remove_headers:
                header.append(value)
            else:
                remove_index.append(x)  # if header is to be removed its column number is stored here
        header = '<tr><th>' + '</th>\n<th>'.join(header) + '</th></tr>'  # html table headers are formatted here
        for starting in starting_points:
            if starting + self.CreateSubPagesForEachXXEntries < output_xlsx.nrows:  # if it is not last sub page
                ending = starting + self.CreateSubPagesForEachXXEntries - 1  # ending point is only used for title
            else:
                ending = output_xlsx.nrows
            title = ((len(str(output_xlsx.nrows)) - len(str(starting))) * "0") + str(starting
                    ) + " - " + ((len(str(output_xlsx.nrows)) - len(str(ending))) * "0") + str(ending
                    ) + " " + self.pageTitle  # generating title for subpage
            data = []
            for row in range(starting, ending):
                dt = []
                for column in range(output_xlsx.ncols):
                    value = output_xlsx.cell_value(row, column)
                    if type(value) == float:  # coverting int and float into string
                        if repr(value)[-2:] == '.0':
                            value = int(value)
                    value = str(value)
                    if column not in remove_index:  # if column is not in remove_header list than add the data in html
                        dt.append(escape(value))
                data.append('<td>' + '</td>\n<td>'.join(dt) + '</td>')  # generating html table row and appending it in list
            # html table row tag is added for every row which are stored in list
            html = '<table><tbody>' + header + '<tr>' + '</tr>\n<tr>'.join(data) + '</tr>' + '</tbody></table>'
            confluence.create_page(
                self.space, title, html, parent_id=parent_page, type='page', representation='storage'
            )  # subpage is created here with the help of parent page id




import csv
import itertools
import xlsxwriter
import xl2dict
import xlrd3 as xlrd
import errno
import os
import logging
import faker
from random import sample, randint
import baangt.base.GlobalConstants as GC
import re

logger = logging.getLogger("pyC")


class TestDataGenerator:
    """
    TestDataGenerator Class is to used to create a TestData file from raw excel file containing all possible values.

    Formats accepted in input excel file:
    1. Value             = ``<value>``
    2. list of values    = ``[<value1>,<value2>]``
    3. range             = ``<start>-<end>,<step>``
    4. random            = ``RND_[list]``
    5. random from range = ``RND_<start>-<end>,<step>``
    6. List of header    = ``[<title1>, <title2>, <title3>]``
    7. Faker Prefix      = ``FKR_(<type>, <locale>, <number_of_data>)``
    8. RRD Prefix        = ``RRD_(<sheetName>,<TargetData>,[<Header1>:[<Value1>],<Header2>:[<Value1>,<Value2>]])``

    :param rawExcelPath: Takes input path for xlsx file containing input data.
    :param sheetName: Name of sheet where all base data is located.
    :method write: Will write the final processed data in excel/csv file.
    """
    def __init__(self, rawExcelPath=GC.TESTDATAGENERATOR_INPUTFILE, sheetName=""):
        self.path = os.path.abspath(rawExcelPath)
        self.sheet_name = sheetName
        if not os.path.isfile(self.path):
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), self.path)
        self.sheet_dict, self.raw_data_json = self.__read_excel(self.path, self.sheet_name)
        self.remove_header = []
        self.processed_datas = self.__process_data(self.raw_data_json)
        self.headers = [x for x in list(self.processed_datas[0].keys()) if x not in self.remove_header]
        self.final_data = self.__generateFinalData(self.processed_datas)

    def write(self, OutputFormat=GC.TESTDATAGENERATOR_OUTPUT_FORMAT, batch_size=0, outputfile=None):
        """
        Will write the generated data in output file.
        :param OutputFormat: "xlsx" or "csv"
        :param batch_size: Number of data to be written in output file. Will be randomly selected.
        :param outputfile: name and path of outputfile.
        :return:
        """
        if OutputFormat.lower() == "xlsx":
            if outputfile == None:
                outputfile = GC.TESTDATAGENERATOR_OUTPUTFILE_XLSX
            self.__write_excel(batch_size=batch_size, outputfile=outputfile)
        elif OutputFormat.lower() == "csv":
            if outputfile == None:
                outputfile = GC.TESTDATAGENERATOR_OUTPUTFILE_CSV
            self.__write_csv(batch_size=batch_size, outputfile=outputfile)
        else:
            logger.debug("Incorrect file format")

    def __write_excel(self, outputfile=GC.TESTDATAGENERATOR_OUTPUTFILE_XLSX, batch_size=0):
        """
        Writes TestData file with final processsed data.
        :param outputfile: Name and path for output file.
        :param batch_size: No. of data to be randomly selected and written in output file.
        :return: None
        """
        if batch_size > 0:
            if len(self.final_data) > batch_size:
                data_lis = sample(self.final_data, batch_size)
            else:
                data_lis = self.final_data
                logger.debug("Total final data is smaller than batch size.")
        else:
            data_lis = self.final_data
        with xlsxwriter.Workbook(outputfile) as workbook:
            worksheet = workbook.add_worksheet()
            worksheet.write_row(0, 0, self.headers)
            for row_num, data in enumerate(data_lis):
                worksheet.write_row(row_num+1, 0, data)

    def __write_csv(self, outputfile=GC.TESTDATAGENERATOR_OUTPUTFILE_CSV, batch_size=0):
        """
        Writes final data in csv
        :param outputfile: Name and path of output file
        :param batch_size: No. of data to be randomly selected and written in output file.
        :return:
        """
        if batch_size > 0:
            if len(self.final_data) > batch_size:
                data_lis = sample(self.final_data, batch_size)
            else:
                data_lis = self.final_data
        else:
            data_lis = self.final_data
        with open(outputfile, 'w', newline='\n', encoding='utf-8-sig') as file:
            fl = csv.writer(file)
            fl.writerow(self.headers)
            for dt in data_lis:
                fl.writerow(list(dt))

    def __generateFinalData(self, processed_data):
        """
        This method will do the final process on the processed_data. Processed_data contains list of dictionary, each
        dictionary is the row from input file which are processed to be interact able in python as per the requirement.

        First loop is of processed_data
        Second loop is of the dictionary(row) and each key:value of that dictionary is header:processed_data

        Method will first check the data type of value.
        If it is a string than method will put it inside a list(i.e. ["string"])
        If it is a tuple than it is a data with prefix so it will be sent to ``__prefix_data_processing`` method for
        further processing.
        Else the value is of type list.

        Then we store all this lists in a list(can be treat as row). This list contains value of cells. They are evaluted
        i.e. ranges are converted to list, strings are converted to list & list are all ready list. So to generate all
        possible combinations from it we use ``iterable`` module.

        Once this list of lists which contains all possible combinations is created we will call
        ``__update_prefix_data_in_final_list`` method. This method will insert the processed prefix data along with the
        data of all combinations list in the final list with the correct position of every value.

        Finally it will return the list of lists which is completely processed and ready to be written in output file.

        :param processed_data:
        :return: Final_data_list
        """
        final_data = []
        for lis in processed_data:
            index = {}
            data_lis = []
            for key in lis:
                if type(lis[key]) == str:
                    data = [lis[key]]
                elif type(lis[key]) == tuple:
                    if len(lis[key]) > 0:
                        self.__prefix_data_processing(lis, key, index)
                        continue
                    else:
                        data = ['']
                else:
                    data = lis[key]
                data_lis.append(data)
            datas = list(itertools.product(*data_lis))
            self.__update_prefix_data_in_final_list(datas, index, final_data)
        logger.info(f"Total generated data = {len(final_data)}")
        return final_data

    def __update_prefix_data_in_final_list(self, data_list, dictionary, final_list):
        """
        This method will insert the data from the dictionary to final_list. So further it can be written in the output.

        ``data_list`` is the list where all possible combinations generated by the input lists and ranges are stored.
        ``dictionary`` is where the data with prefix are stored with their index value(which will be used to place data)
        ``final_list`` is the list where final data will be stored after merging values from data_list and dictionary

        First it will iterate through data_list which is list of lists(Also you can take it as list of rows).
        Second loop will go through dictionary and check the data type of each value. Pairings inside dictionary are of
        ``index: value`` here index is the position from where this value was picked. So it will be used in placing the
        data in their correct position.

        List:
        =====
            If value is list then it is a data with ``FKR_`` prefix with number of data 0(i.e. create new fake data for
            every output). So we will create the faker module instance as per the input and will generate fake data for
            every row. And will insert them in the position of index(key of dictionary).


        Dictionary:
        ==========
            If it is not of type list then we will check that if it is of type dict. If yes then this is a data with
            "RRD_" prefix and we have to select the random data here. So we will start looping through this dictionary.
            Remember this is the third loop. This dictionary contains header:value(TargetDatas from matched data) pair.
            On every loop it will first check that if the same header is stored in done dictionary. If yes then it will
            get value of it from done dictionary. Then it will create a list from the TargetData list. This new list
            will contain only data which has same value for same header stored in done dictionary.
            i.e. If matching Header has value x then from TargetDatas list only data where header=x will be
            considered for random pick.

            Then the random value is selected from that list.
            If none of the header is processed before(for the same row). Then it will get random data from the list and
            will store header: value pair in done dictionary so it is used in the checking process as above.

            It will also check if the ``self.header`` list contains all the header which are in the random selected data.
            If not then it will add the header there.

            At last we will index the position of header inside self.headers list and will insert the value in same
            position and append the row in final_data list.

        Tuple:
        ======
            If the type is tuple then we need to simply pick a random value from it and insert it in the same position of
            index(key of dictionary for current value).

        :param data_list:
        :param dictionary:
        :param final_list:
        :return: None
        """
        for data in data_list:
            data = list(data)
            done = {}
            for ind in dictionary:
                if type(dictionary[ind]) == list:
                    fake = faker.Faker(dictionary[ind][2])
                    data.insert(ind, getattr(fake, dictionary[ind][1])())
                else:
                    if type(dictionary[ind][0]) == dict:
                        sorted_data = False
                        for header in dictionary[ind][0]:
                            if header in done:
                                match = done[header]
                                sorted_data = [x for x in dictionary[ind] if x[header] == match]
                                break
                        if not sorted_data:
                            sorted_data = dictionary[ind]
                        data_to_insert = sorted_data[randint(0, len(sorted_data) - 1)]
                        for keys in data_to_insert:
                            if keys not in self.headers:
                                self.headers.append(keys)
                            if keys not in done:
                                data.insert(self.headers.index(keys), data_to_insert[keys])
                                done[keys] = data_to_insert[keys]
                    else:
                        data_to_insert = dictionary[ind][randint(0, len(dictionary[ind]) - 1)]
                        data.insert(ind, data_to_insert)
            final_list.append(data)

    def __prefix_data_processing(self, dic, key, dictionary: dict):
        """
        This method will process the datas with prefix.

        ``dic`` the dictionary where all data which are in final process is stored
        ``key`` the header of the current data which will be used now to call the data.
        ``dictionary`` in which the values will be inserted after performing their process.

        First it will check the first value of tuple.
        If it is ``Faker`` then in will continue the process and will check the 4th value of tuple.
        If the 4th value(which is used to determine the number of fake data to be generated and store inside a list)
        is ``0`` then the method will store the values as it is in a list, because ``0`` value means we have to generate
        new fake data for every output data, so it will be done later.
        If it is greater than ``0`` then this method will create tuple with the given number of fake data and store it.
        (If no number is given then default number is 5.)

        If first value is not ``Faker`` then no process will be done.

        Finally the data will be inserted in the dictionary.

        :param dic:
        :param key:
        :param dictionary:
        :return:
        """
        ltuple = dic[key]
        if ltuple[0] == "Faker":
            fake = faker.Faker(ltuple[2])
            fake_lis = []
            if len(ltuple) == 4:
                if int(ltuple[3]) == 0:
                    dictionary[list(dic.keys()).index(key)] = list(ltuple)
                    return True
                else:
                    for x in range(int(ltuple[3])):
                        fake_lis.append(getattr(fake, ltuple[1])())
            else:
                for x in range(5):
                    fake_lis.append(getattr(fake, ltuple[1])())
            dictionary[list(dic.keys()).index(key)] = tuple(fake_lis)
            return True
        else:
            dictionary[list(dic.keys()).index(key)] = ltuple
            return True

    def __process_data(self, raw_json):
        """
        This method is used to Process all the raw unprocessed data read from the excel file.

        It will first send the header to ``__data_generator`` so that if it is a list then it will get converted in
        individual header.

        Later it will process the values using ``__data_generator``.

        It will then check returned iterable type, if it is a tuple that mean input value was with prefix, so, it will
        further check if the tuple contains dict. If True than prefix was RRD_. In that case we will have to deal with
        the original header of the input value. Because if the original value's header is not in the TargetData then this
        header will contain no value in the output file and my cause errors too. So the header will added in
        ``self.remove_header`` list which will be further used to remove it from main header list.

        Finally it will return list of dictionarys. Each dictionary contains processed data of a row of input file.
        Processed data are the raw data converted into python data type and iterables. Ranges are converted into list.

        :param raw_json:
        :return:
        """
        processed_datas = []
        for raw_data in raw_json:
            if not list(raw_data.values())[0]:
                continue
            processed_data = {}
            for key in raw_data:
                keys = self.__data_generators(key)
                for ke in keys:
                    processed_data[ke] = self.__data_generators(raw_data[key])
                    if type(processed_data[ke]) == tuple and len(processed_data[ke])>0:
                        if type(processed_data[ke][0]) == dict:
                            if ke not in processed_data[ke][0]:
                                self.remove_header.append(ke)
            processed_datas.append(processed_data)
        return processed_datas

    def __data_generators(self, raw_data):
        """
        This method first send the data to ``__raw_data_string_process`` method to split the data and remove the unwanted
        spaces.

        Later this method uses other methods to convert all the different data_types from string to their respective
        python data types.
        i.e. string list to python list, etc.

        Later according to the prefix of data and the data_type assigned it will convert them.
        Simple list and strings are converted in to ``list`` type.
        Data with prefix will converted in to ``tuple`` type so further it will be helpful in distinguishing. Also it will
        insert the prefix name in first value of tuple if the prefix is ``FKR_`` so it will be helpful in further process.

        Finally it will return the iterable for further process.

        :param raw_data:
        :return: List or Tuple containing necessary data
        """
        raw_data, prefix, data_type = self.__raw_data_string_process(raw_data)
        if len(raw_data)<=1:
            return [""]
        if raw_data[0] == "[" and raw_data[-1] == "]" and prefix == "":
            processed_datas = self.__splitList(raw_data)
            processed_datas = data_type(processed_datas)

        elif prefix == "Faker":
                processed_datas = [data.strip() for data in raw_data[1:-1].split(",")]
                processed_datas.insert(0, "Faker")
                processed_datas = data_type(processed_datas)

        elif prefix == "Rrd":
            first_value = raw_data[1:-1].split(',')[0].strip()
            second_value = raw_data[1:-1].split(',')[1].strip()
            if second_value[0] == "[":
                second_value = ','.join(raw_data[1:-1].split(',')[1:]).strip()
                second_value = second_value[:second_value.index(']')+1]
                third_value = [x.strip() for x in ']'.join(raw_data[1:-1].split(']')[1:]).split(',')[1:]]
            else:
                third_value = [x.strip() for x in raw_data[1:-1].split(',')[2:]]
            evaluated_list = ']],'.join(','.join(third_value)[1:-1].strip().split('],')).split('],')
            if evaluated_list[0] == "":
                evaluated_dict = {}
            else:
                evaluated_dict = {
                    splited_data.split(':')[0]: self.__splitList(splited_data.split(':')[1])  for splited_data in evaluated_list
                }
            if second_value[0] == "[" and second_value[-1] == "]":
                second_value = self.__splitList(second_value)
            processed_datas = self.__processRrd(first_value, second_value,evaluated_dict)
            processed_datas = data_type(processed_datas)

        elif prefix == "Rre":
            file_name = raw_data[1:-1].split(',')[0].strip()
            sheet_dict, _ = self.__read_excel(file_name)
            first_value = raw_data[1:-1].split(',')[1].strip()
            second_value = raw_data[1:-1].split(',')[2].strip()
            if second_value[0] == "[":
                second_value = ','.join(raw_data[1:-1].split(',')[2:]).strip()
                second_value = second_value[:second_value.index(']')+1]
                third_value = [x.strip() for x in ']'.join(raw_data[1:-1].split(']')[1:]).split(',')[1:]]
            else:
                third_value = [x.strip() for x in raw_data[1:-1].split(',')[3:]]
            evaluated_list = ']],'.join(','.join(third_value)[1:-1].strip().split('],')).split('],')
            if evaluated_list[0] == "":
                evaluated_dict = {}
            else:
                evaluated_dict = {
                    splited_data.split(':')[0]: self.__splitList(splited_data.split(':')[1])  for splited_data in evaluated_list
                }
            if second_value[0] == "[" and second_value[-1] == "]":
                second_value = self.__splitList(second_value)
            processed_datas = self.__processRrd(first_value, second_value, evaluated_dict, sheet_dict)
            processed_datas = data_type(processed_datas)

        elif "-" in raw_data:
            raw_data = raw_data.split('-')
            start = raw_data[0].strip()
            end = raw_data[1].strip()
            step = 1
            if "," in end:
                raw_data = end.split(",")
                end = raw_data[0].strip()
                step = raw_data[1].strip()
            processed_datas = [x for x in range(int(start), int(end)+1, int(step))]
            processed_datas = data_type(processed_datas)

        else:
            processed_datas = [raw_data.strip()]
            processed_datas = data_type(processed_datas)
        return processed_datas

    def __raw_data_string_process(self, raw_string):
        """
        Returns ``String, prefix, data_type`` which are later used to decided the process to perform on string.
        Their depth explanation are written in the function where they are used.

        It will process the value string of all cells in the input sheet.
        It will first convert all floats into string as by default xlrd ints are converted in float
        Later it will check if the string size is greater than 4 or not. If not then it will simply return the values,
        else it will process further.

        If string has more than 4 characters, this method will look if the fourth character is "_" or not. If not it will
        return the values. Else it mean there is prefix in string and it will process further.

        Later it will split the prefix from the value and define the data_type according to the string.
        If their is no matching prefix then the data type wil be list else it will be tuple.

        :param raw_string:
        :return: String of values, prefix, Data_type
        """
        if type(raw_string) == float:
            raw_string = int(raw_string)
        raw_string = str(raw_string).strip()
        prefix = ""
        if len(raw_string)>4:
            if raw_string[3] == "_":
                if raw_string[:4].lower() == "rnd_":           # Random
                    raw_string = raw_string[4:]
                    data_type = tuple
                elif raw_string[:4].lower() == "fkr_":
                    prefix = "Faker"
                    raw_string = raw_string[4:]
                    data_type = tuple
                elif raw_string[:4].lower() == "rrd_":         # Remote Random (Remote = other sheet)
                    prefix = "Rrd"
                    raw_string = self.__process_rrd_string(raw_string)
                    raw_string = raw_string[4:]
                    data_type = tuple
                elif raw_string[:4].lower() == "rre_":         # Remote Random (Remote = other sheet)
                    prefix = "Rre"
                    raw_string = self.__process_rre_string(raw_string)
                    raw_string = raw_string[4:]
                    data_type = tuple
                else:
                    data_type = list
            else:
                data_type = list
        else:
            data_type = list
        return raw_string, prefix, data_type

    @staticmethod
    def __read_excel(path, sheet_name=""):
        """
        This method will read the input excel file.
        It will read all the sheets inside this excel file and will create a dictionary of dictionary containing all data
        of every sheet.
        i.e. {"sheetName": {headers**: data**}}

        It will also look for a base sheet whose name must be given while creating the instance. If no sheet name is
        given then first sheet of the file will be considered as base sheet.

        Finally it will return a dictionary containing sheetNames:data of all sheets & dictionary of base sheet.

        :param path: Path to raw data xlsx file.
        :param sheet_name: Name of base sheet sheet where main input data is located. Default will be the first sheet.
        :return: Dictionary of all sheets and data, Dictionary of base sheet.
        """
        wb = xlrd.open_workbook(path)
        sheet_lis = wb.sheet_names()
        sheet_dict = {}
        for sheet in sheet_lis:
            xl_obj = xl2dict.XlToDict()
            data = xl_obj.fetch_data_by_column_by_sheet_name(path,sheet_name=sheet)
            sheet_dict[sheet] = data
        if sheet_name == "":
            base_sheet = sheet_dict[sheet_lis[0]]
        else:
            assert sheet_name in sheet_dict, f"Excel file doesn't contain {sheet_name} sheet. Please recheck."
            base_sheet = sheet_dict[sheet_name]
        return sheet_dict, base_sheet

    @staticmethod
    def __splitList(raw_data):
        """
        Will convert string list to python list.
        i.e. "[value1,value2,value3]" ==> ["value1","value2","value3"]
        :param raw_data: string of list
        :return: Python list
        """
        proccesed_datas = [data.strip() for data in raw_data[1:-1].split(",")]
        return proccesed_datas

    def __processRrd(self, sheet_name, data_looking_for, data_to_match: dict, sheet_dict=None, caller="RRD_"):
        """
        This function is internal function to process the data wil RRD_ prefix.
        The General input in excel file is like ``RRD_[sheetName,TargetData,[Header:[values**],Header:[values**]]]``
        So this program will take already have processed input i.e. strings converted as python string and list to
        python list, vice versa.

        ``sheet_name`` is the python string referring to TargetData containing string.

        ``data_looking_for`` is expected to be a string or list of TargetData. When there are multiple values then the
        previous process will send list else it will send string. When a string is received it will be automatically
        converted in list so the program will alwaus have to deal with list. If input is "*" then this method will take
        all value in the matched row as TargetData.

        ``data_to_match`` is a python dictionary created by the previous process. It will contain the key value pair
        same as given in input but just converted in python dict. Then all possible combinations will be generated inside
        this method. If an empty list is given by the user in the excel file then this list will get emptu dictionary from
        the previous process. Thus then this method will pick TargetData from all rows of the target sheet.

        :param sheet_name:
        :param data_looking_for:
        :param data_to_match:
        :return: dictionary of TargetData
        """
        sheet_dict = self.sheet_dict if sheet_dict is None else sheet_dict
        matching_data = [list(x) for x in itertools.product(*[data_to_match[key] for key in data_to_match])]
        assert sheet_name in sheet_dict, \
            f"Excel file doesn't contain {sheet_name} sheet. Please recheck. Called in '{caller}'"
        base_sheet = sheet_dict[sheet_name]
        data_lis = []
        if type(data_looking_for) == str:
            data_looking_for = data_looking_for.split(",")

        for data in base_sheet:
            if len(matching_data) == 1 and len(matching_data[0]) == 0:
                if data_looking_for[0] == "*":
                    data_lis.append(data)
                else:
                    data_lis.append({keys: data[keys] for keys in data_looking_for})
            else:
                if [data[key] for key in data_to_match] in matching_data:
                    if data_looking_for[0] == "*":
                        data_lis.append(data)
                    else:
                        data_lis.append({keys: data[keys] for keys in data_looking_for})
        return data_lis

    def __process_rrd_string(self, rrd_string):
        """
        This method is used to validate rrd_strings provided by the user.
        If their will be any error in string this fuction will immediately create an error and will stop further execution.
        Also these function will remove empty spaces around the commas in string.
        Regex supporting formats in this method are:
        ``RRD_[sheetName,TargetData,[Header:[values**],Header:[values**]]]``
        ``RRD_[sheetName,[TargetData**],[Header:[values**],Header:[values**]]]``
        ``RRD_(sheetName,[TargetData**],[Header:[values**],Header:[values**]])``
        ``RRD_[sheetName,*,[Header:[values**],Header:[values**]]]``
        ``RRD_[sheetName,*,[Header:[values**],Header:[values**]]]``
        ``RRD_[sheetName,TargetData,[]]``
        ``RRD_(sheetName,TargetData,[])``
        ``RRD_(sheetName,*,[])``
        ``RRD_[sheetName,*,[]]``
        :param rrd_string:
        :return:
        """
        processed_string = ','.join([word.strip() for word in rrd_string.split(', ')])
        match = re.match(r"(RRD_(\(|\[))[a-zA-z0-9\s]+,(\[?[a-zA-z\s,]+\]?|)|\*,\[([a-zA-z0-9\s]+:\[[a-zA-z0-9,\s]+\](,?))*\]",processed_string)
        err_string = f"{rrd_string} not matching pattern RRD_(sheetName,TargetData," \
                     f"[Header1:[Value1],Header2:[Value1,Value2]])"
        assert match, err_string
        return processed_string

    def __process_rre_string(self, rrd_string):
        """
        This method is used to validate rrd_strings provided by the user.
        If their will be any error in string this fuction will immediately create an error and will stop further execution.
        Also these function will remove empty spaces around the commas in string.
        Regex supporting formats in this method are:
        ``RRE_[fileName,sheetName,TargetData,[Header:[values**],Header:[values**]]]``
        ``RRE_[fileName,sheetName,[TargetData**],[Header:[values**],Header:[values**]]]``
        ``RRE_(fileName,sheetName,[TargetData**],[Header:[values**],Header:[values**]])``
        ``RRE_[fileName,sheetName,*,[Header:[values**],Header:[values**]]]``
        ``RRE_[fileName,sheetName,*,[Header:[values**],Header:[values**]]]``
        ``RRE_[fileName,sheetName,TargetData,[]]``
        ``RRE_(fileName,sheetName,TargetData,[])``
        ``RRE_(fileName,sheetName,*,[])``
        ``RRE_[fileName,sheetName,*,[]]``
        :param rrd_string:
        :return:
        """
        processed_string = ','.join([word.strip() for word in rrd_string.split(', ')])
        match = re.match(r"(RRE_(\(|\[))[\w\d\s\-./\\]+\.(xlsx|xls),[a-zA-z0-9\s]+,(\[?[a-zA-z\s,]+\]?|)|\*,\[([a-zA-z0-9\s]+:\[[a-zA-z0-9,\s]+\](,?))*\]",processed_string)
        err_string = f"{rrd_string} not matching pattern RRE_(fileName, sheetName, TargetData," \
                     f"[Header1:[Value1],Header2:[Value1,Value2]])"
        assert match, err_string
        return processed_string

if __name__ == "__main__":
    lTestDataGenerator = TestDataGenerator("../../tests/0TestInput/RawTestData.xlsx")
    lTestDataGenerator.write()



from baangt.base.BrowserHandling.BrowserHandling import BrowserDriver
from baangt.base.ProxyRotate import ProxyRotate
from baangt.base.Utils import utils
import baangt.base.GlobalConstants as GC

import logging
from os import getcwd
from time import sleep

logger = logging.getLogger("pyC")


class BrowserFactory:
    """
    This is a class, that holds the interface to browser-sessions.

    """
    def __init__(self, testrun):
        self.testrun = testrun

        self.rotatingProxiesService = None
        self.browser = {}
        self.browserInstances = {}

        self.timing = self.testrun.timing
        self.globalSettings = testrun.globalSettings
        self.browserServer = BrowserFactory.__getBrowserServer() \
            if self.globalSettings.get('TC.' + GC.EXECUTION_NETWORK_INFO) == True else None
        self.browsersMobProxies = {}

        self.__startRotatingProxies()

    def __startRotatingProxies(self):
        if self.globalSettings.get("TC.UseRotatingProxies"):
            reReadProxies = False
            if self.globalSettings.get("TC.ReReadProxies"):
                reReadProxies = True
            self.rotatingProxiesService = ProxyRotate(reReadProxies=reReadProxies)
            if self.globalSettings.get("TC.ReReadProxies"):
                self.rotatingProxiesService.recheckProxies(forever=True)

    def getBrowser(self, browserInstance=0, browserName=None, browserAttributes=None,
                         mobileType=None, mobileApp=None, desired_app=None,
                         mobile_app_setting=None, browserWindowSize = None):
        """
                This method is called whenever a browser instance (existing or new) is needed. If called without
                parameters it will create one instance of Firefox (geckodriver).

                if global setting TC.EXECUTION_SLOW is set, inform the browser instance about it.

                @param browserInstance: Number of the requested browser instance. If none is provided, always the
                                        default browser instance 0 will be returned
                @param browserName: one of the browser names (e.g. FF, Chrome) from GC.BROWSER*
                @param browserAttributes: optional Browser Attributes
                @return: the browser instance of base class BrowserDriver

                """
        if mobileType == 'True':
            logger.info(f"opening new Appium instance {browserInstance} of Appium browser {browserName}")
            self._getBrowserInstance(browserInstance=browserInstance)
            self.setBrowserProxy(browserInstance=browserInstance)
            if self.browsersMobProxies:
                browserMobProxy = self.browsersMobProxies[browserInstance]
            else:
                browserMobProxy = None
            self.browser[browserInstance].createNewBrowser(mobileType=mobileType,
                                                           mobileApp=mobileApp,
                                                           desired_app=desired_app,
                                                           mobile_app_setting=mobile_app_setting,
                                                           browserName=browserName,
                                                           desiredCapabilities=browserAttributes,
                                                           browserProxy=browserMobProxy,
                                                           browserInstance=browserInstance)
            if self.globalSettings.get("TC." + GC.EXECUTION_SLOW):
                self.browser[browserInstance].slowExecutionToggle()
            return self.browser[browserInstance]
        else:
            if self.globalSettings.get("TC.RestartBrowser"):
                if browserInstance in self.browser.keys():
                    logger.debug(f"Instance {browserInstance}: TC.RestartBrowser was set. Quitting old browser.")
                    lBrowser = self.browser[browserInstance]
                    lBrowser.closeBrowser()
                    del self.browser[browserInstance]

            if browserInstance not in self.browser.keys():
                logger.info(f"opening new instance {browserInstance} of browser {browserName}")
                self._getBrowserInstance(browserInstance=browserInstance)
                self.setBrowserProxy(browserInstance=browserInstance)
                if self.browsersMobProxies:  # Locale call of browserMob-Proxy
                    browserMobProxy = self.browsersMobProxies[browserInstance]
                else:
                    browserMobProxy = None

                randomProxy = None
                if self.globalSettings.get("TC.UseRotatingProxies"):
                    # returns a Dict of ip and port
                    randomProxy = self.rotatingProxiesService.random_proxy()

                self.browser[browserInstance].createNewBrowser(mobileType=mobileType,
                                                               mobileApp=mobileApp,
                                                               desired_app=desired_app,
                                                               mobile_app_setting=mobile_app_setting,
                                                               browserName=browserName,
                                                               desiredCapabilities=browserAttributes,
                                                               browserProxy=browserMobProxy,
                                                               browserInstance=browserInstance,
                                                               randomProxy=randomProxy)

                if self.globalSettings.get("TC." + GC.EXECUTION_SLOW):
                    self.browser[browserInstance].slowExecutionToggle()
                if browserWindowSize:
                    self.setBrowserWindowSize(self.browser[browserInstance], browserWindowSize)
                if self.globalSettings.get("TC." + GC.BROWSER_ZOOM_FACTOR):
                    self.browser[browserInstance].setZoomFactor(self.globalSettings["TC." + GC.BROWSER_ZOOM_FACTOR])
            else:
                logger.debug(f"Using existing instance of browser {browserInstance}")
            return self.browser[browserInstance]

    @staticmethod
    def setBrowserWindowSize(lBrowserInstance: BrowserDriver, browserWindowSize):
        lBrowserInstance.setBrowserWindowSize(browserWindowSize)

    def _getBrowserInstance(self, browserInstance):
        if self.testrun.classesForObjects.browserHandling:
            lClass = utils.dynamicImportOfClasses(fullQualifiedImportName=self.testrun.classesForObjects.browserHandling)

            self.browser[browserInstance] = lClass(timing=self.timing)
        else:
            # !Sic: code duplication for convenince reasons. Pure Duck-Typing would prevent where-used-list to work.
            self.browser[browserInstance] = BrowserDriver(timing=self.timing)

    @staticmethod
    def __getBrowserServer():
        from browsermobproxy import Server
        server = Server(getcwd() + GC.BROWSER_PROXY_PATH)
        logger.info("Starting browsermob proxy")
        server.start()
        return server

    def setBrowserProxy(self, browserInstance):

        sleep(1)

        proxy = self.browserServer.create_proxy() if self.browserServer else None

        if not proxy:
            return

        sleep(1)

        self.browsersMobProxies[browserInstance] = proxy

    def teardown(self):
        network_info = {}
        if not self.globalSettings.get("TC." + GC.EXECUTION_DONTCLOSEBROWSER):
            for browserInstance in self.browser.keys():
                self.browser[browserInstance].closeBrowser()

        elif self.globalSettings.get("TC." + GC.EXECUTION_DONTCLOSEBROWSER) == "False":
            for browserInstance in self.browser.keys():
                self.browser[browserInstance].closeBrowser()

        if self.browserServer:
            network_info = [info.har if info else {} for info in self.browsersMobProxies.values()]
            self.browserServer.stop()

        return network_info
